/*
 *  This file is part of Dascant.
 *
 *  Dascant is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Dascant is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Dascant; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
 *  USA
 */

/**
 * The camera class.  A camera keeps track of its own rays, for the purpose
 * of raytracing.  Repositioning the camera causes new rays to be generated
 * (but not traced).
 *
 * The results are stored in a one-dimensional unsigned charater array, as a
 * sequence of RGB-values, going across and then down (row-major order, I
 * think).
 */

#include <stdio.h>
#include <camera.h>
#include <collision.h>
#include <string.h>

/**
 * Class constructor for camera with default values.
 */
camera::
camera(void)
{
	initialize(128, 128, 200);
}

/**
 * Class constructor.
 * @param w width of the image plane
 * @param h height of the image plane
 * @param d image plane depth (distance from the camera center).
 */
camera::
camera( int w, int h, double d )
{
	initialize(w, h, d);
}

void camera::
initialize( int w, int h, double d )
{
	width = w;
	height = h;
	depth = d;

	orientation.setAngle(0);

	rays = new (ray *)[height];
	for (int i = 0; i < height; ++i)
		rays[i] = new ray[width];
	pixel = new unsigned char[height*width*3];

	updatePosition = false;
	updateOrientation = false;
}

camera::
~camera()
{
	delete pixel;

	for (int i = 0; i < height; ++i)
		delete rays[i];
	delete rays;
}

/**
 * Create a copy of this camera.
 * @return a pointer to a new camera object.
 */
camera *camera::
copy()
{
	camera *newCam = new camera(width, height, depth);
	newCam->position = position;
	newCam->height = height;
	newCam->width = width;
	//newCam.ray = new ray[height][width];
	for (int i = 0; i < height; ++i)
		for (int j = 0; j < width; ++j)
			newCam->rays[i][j] = rays[i][j];
	return newCam;
}

/**
 * Set this camera's position in space, recompute all initial rays.  Specify
 * position as four doubles.
 */
void camera::
setPosition( double x, double y, double z, double w )
{
	position(x, y, z, w);
	compute_rays();
}

/**
 * Set this camera's position and orientation in space, recomputing all
 * initial rays.  Specify a vector for position and a rotation for
 * orientation.
 */
void camera::
setPosition( const vector4 *V, const rotation4 *R )
{
	position = *V;
	orientation = *R;
	compute_rays();
}

/**
 * Set the next position of the camera.  FIXME: this isn't used anymore.
 * @param V the next position.
 */
void camera::
enqueuePosition( const vector4 *V )
{
	nextPosition = *V;
	updatePosition = true;
}

/**
 * Set the next orientation for the camera.  FIXME: also not used anymore.
 * @param R the next rotation.
 */
void camera::
enqueueOrientation( const rotation4 *R )
{
	nextOrientation = *R;
	updateOrientation = true;
}

/**
 * Precompute the rays.
 */
void camera::
compute_rays(void)
{
	int i, j;

	for (i = 0; i < height; ++i)
		for (j = 0; j < width; ++j)
			rays[i][j](this, j-width/2, i-height/2);
}

/**
 * Trace all the rays and see if they hit an entity.  Set the pixel array
 * accordingly.  This is the heart of the raytracing algorithm, and the part
 * that probably needs to be expanded a bit if we want to actually allow
 * rays to bounce off of stuff.
 * @param E the entity for checking collisions against.
 */
void camera::
snapshot( const entity *E )
{
	int i, j, k, nColl;
	collision *C;

	C = new collision[2*E->size()];

	for (i = 0; i < height; ++i)
		for (j = 0; j < width; ++j) {
			/* blank out the current pixel */
			for (k = 0; k < 3; ++k)
				pixel[i*3*width + j*3 + k] = 0;

			/* check for collisions */
			nColl = E->detectCollisions(&rays[height-i-1][j], C);
			if (nColl == 0) continue;

			/* figure out the closest collision */
			if (C[0].rayPosition > 0)
				k = 0;
			else if (nColl > 1 && C[1].rayPosition > 0)
				k = 1;
			else
				continue;

			/* compute the normal, which also computes the color */
			C[k].computeNormal();
			pixel[i*width*3 + j*3 + 0] = C[k].red;
			pixel[i*width*3 + j*3 + 1] = C[k].green;
			pixel[i*width*3 + j*3 + 2] = C[k].blue;
		}

	delete C;
}
