/*
 *  This file is part of Dascant.
 *
 *  Dascant is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Dascant is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Dascant; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
 *  USA
 */

/**
 * The collision class.
 */

#include <collision.h>
#include <stdio.h>
#include <dascant.h>

void collision::
initialize( const entity *e )
{
	inside = false;
	object = e;
	rayPosition = 0;
}

/**
 * Class constructor with ray.
 * @param R the ray causing the collision.
 * @param t the distance along the ray that the collision occured.
 * @param e the object the ray struck.
 */
collision::
collision( const ray *R, double t, const entity *e )
{
	initialize(e);
	location = R->evalAt(t);
	rayPosition = t;
}

/**
 * Fill an array with a string representation of this collision.  If str is
 * not specified (or is NULL), a static buffer will be used.
 * @param str a buffer to fill, or NULL; defaults to NULL.
 * @return a pointer to the written string.
 */
char *collision::
toString( char *str )
{
	char b1[256], buf[256];

	if (!str) str = buf;

	location.toString(b1);
	sprintf(str, "[ %s, t=%g ]", b1, rayPosition);

	return str;
}

/**
 * Compare two collisions.  C-like comparison; closer collisions are
 * "lesser".
 */
int
collisionCmp( const collision *a, const collision *b )
{
	return a->rayPosition > b->rayPosition;
#if 0
	if (a->rayPosition > b->rayPosition)
		return 1;
	else if (b->rayPosition > a->rayPosition)
		return -1;
	else
		return 0;
#endif
}

/**
 * Compare collision pointers.
 */
int
collisionPtrCmp( const collision **a, const collision **b )
{
	return collisionCmp(*a, *b);
}

bool collision::operator
<( collision &other )
{
	return rayPosition < other.rayPosition;
}

/**
 * Recompute this collision
 * @param R the ray causing the collision.
 * @param t the distance along the ray from the camera center.
 * @param e the entity being collided at.
 */
collision &collision::operator
()( const ray *R, double t, const entity *e )
{
	collision C(R, t, e);
	*this = C;
	return *this;
}

/**
 * Dummy constructor.  Should not be used by most callers; maybe should be
 * private or protected or something.
 */
collision::
collision( void )
{
}

/**
 * Compute the normal vector to the surface at the point of collision.  Also
 * computes the color that the surface will appear (taking into account the
 * angle to the light source)
 */
void collision::
computeNormal( void )
{
	double sq[4], cr[4][4], li[4];
	int j;

	/* square terms */
	sq[0] = surface->coeff[2][0][0][0];
	sq[1] = surface->coeff[0][2][0][0];
	sq[2] = surface->coeff[0][0][2][0];
	sq[3] = surface->coeff[0][0][0][2];

	/* cross terms */
	cr[0][1] = surface->coeff[1][1][0][0];
	cr[0][2] = surface->coeff[1][0][1][0];
	cr[0][3] = surface->coeff[1][0][0][1];
	cr[1][2] = surface->coeff[0][1][1][0];
	cr[1][3] = surface->coeff[0][1][0][1];
	cr[2][3] = surface->coeff[0][0][1][1];

	/* linear terms */
	li[0] = surface->coeff[1][0][0][0];
	li[1] = surface->coeff[0][1][0][0];
	li[2] = surface->coeff[0][0][1][0];
	li[3] = surface->coeff[0][0][0][1];

	for (j = 0; j < 4; ++j) {
		// consider square term
		normal[j] = 2 * location[j] * sq[j];
		// consider cross terms
		for (int k = 0; k < 4; ++k)
			if (j < k)
				normal[j] += location[k] * cr[j][k];
			else if (j > k)
				normal[j] += location[k] * cr[k][j];
		// consider linear term
		normal[j] += li[j];
	}
	normal.normalize();

	/* color of reflection */
	double cosine = normal.dot(rcfile::lightSource);
	if (cosine <= 0) {
		red = 0;
		green = 0;
		blue = 0;
		return;
	}

	red =   (unsigned char)(surface->red   * cosine);
	green = (unsigned char)(surface->green * cosine);
	blue =  (unsigned char)(surface->blue  * cosine);
}
