/*
 *  This file is part of Dascant.
 *
 *  Dascant is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Dascant is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Dascant; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
 *  USA
 */

/**
 * The frame class.  Represents a local frame of reference or coordinate
 * system.  Pretty useful, at least for cameras.
 */
#include <frame.h>
#include <stdio.h>
#include <rotation4.h>
#include <config.h>
#include <rcfile.h>
#include <math.h>

/**
 * Default constructor.
 */
frame::
frame( void )
{
	/* Create the canonical coordinate system. */
	origin(0, 0, 0, 0);
	for (int i = 0; i < 4; ++i)
		for (int j = 0; j < 4; ++j)
			local[i][j] = i == j;

	if (ran_setup) return;

	/* precompute rotations around local axes */
	ran_setup = true;

	vector4 axis[4];
	rotation4 *R;

	axis[0](1,0,0,0);
	axis[1](0,1,0,0);
	axis[2](0,0,1,0);
	axis[3](0,0,0,1);

	for (int u = 0; u < 4; ++u)
		for (int v = u+1; v < 4; ++v) {
			R = new rotation4(1<<u, 1<<v);
			R->setAngle(rcfile::rotate_stepsize);

			for (int n = 0; n < 4; ++n)
				image[u][v][0][n] = R->applyTo(axis[n]);
			R->setAngle(-rcfile::rotate_stepsize);
			for (int n = 0; n < 4; ++n)
				image[u][v][1][n] = R->applyTo(axis[n]);

			delete R;
		}
}

/** 
 * Translate this frame by v.
 */
void frame::
move( const vector4 &v )
{
	for (int i = 0; i < 4; ++i)
		origin[i] += v.elem[i];
}

/**
 * Rotate this frame of reference around the local uv-plane.
 * @param u 0, 1, 2, or 3 to indicate x, y, z, or w.
 * @param v 0, 1, 2, or 3 to indicate x, y, z, or w.
 * @param s 1 to indicate a reverse rotation, defaults to 0.
 */
void frame::
rotate( int u, int v, int s )
{
	vector4 newAxis[4];

	s = !!s;

	/* Take the precomputed rotated axes, and represent them in the absolute
	 * coordinate system, pretending that this frame is at the standard
	 * origin. */
	for (int i = 0; i < 4; ++i) {
		newAxis[i] = toAbsoluteNoTranslation(image[u][v][s][i]);
		newAxis[i].normalize();
	}

	/* Make the axes of this frame the axes as computed above. */
	for (int i = 0; i < 4; ++i)
		local[i] = newAxis[i];
}

/**
 * Represent a vector in this frame in the standard frame, assuming this
 * frame to be at the origin.  Does not take into account translation of
 * this frame; only rotations are important.
 * @param u a vector represented in this frame.
 * @return a vector in the absolute frame.
 */
vector4 frame::
toAbsoluteNoTranslation( const vector4 &u )
{
	vector4 v;
	int i, j;

	for (i = 0; i < 4; ++i)
		for (j = 0; j < 4; ++j)
			v[i] += local[j][i] * u.elem[j];

	return v;
}

/**
 * Represent a vector in this frame in the standard frame.
 */
vector4 frame::
toAbsolute( const vector4 &u )
{
	vector4 v;

	v = toAbsoluteNoTranslation(u);
	return v + origin;
}

/** 
 * Represent a vector from the absolute frame in this frame.
 */
vector4 frame::
fromAbsolute( const vector4 &v )
{
	vector4 u, V=v;

	for (int i = 0; i < 4; ++i)
		V[i] -= origin[i];

	for (int i = 0; i < 4; ++i)
		u[i] += V.dot(local[i]);

	return u;
}

bool frame::ran_setup = false;

vector4 frame::image[4][4][2][4];
