/*
 *  This file is part of Dascant.
 *
 *  Dascant is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Dascant is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Dascant; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
 *  USA
 */

/**
 * Glue.  This class is used to stick several entities together.  The
 * sticking can be union-like (in which case a collision with any object
 * will be considered collision with the glue) or intersection-like (in
 * which case a ray must collide with the intersection of the solids defined
 * by the quadric surfaces in the glue).
 */
#include <glue.h>
#include <collision.h>
#include <stdlib.h>
#include <stdio.h>

/** 
 * Constructor for a glue that can hold up to n entities.
 */
glue::
glue( int n )
	: entity()
{
	component = new (entity *)[n];
	capacity = n;
	num_components = 0;
	for (int i = 0; i < n; ++i)
		component[i] = 0x0;
	combine_mode = UNION;
}

glue::
~glue()
{
	delete component;
}

/** 
 * Return the number of atomic (non-glue) entities in this glue, counting
 * recursively.
 */
int glue::
size( void ) const
{
	int s = 0;
	for (int i = 0; i < num_components; ++i)
		s += component[i]->size();
	return s;
}

/** 
 * Stick a new entity into this glue.
 */
int glue::
add( entity *e )
{
	if (num_components == capacity)
		return 0;
	component[num_components++] = e;
	return 1;
}

/**
 * Detect collisions with this glue.  As stated, collision detection can be
 * performed in two ways:
 * 1. UNION mode: A collision with any surface is sufficient for a collision
 * with the glue; the glue is considered the union of its components.
 * 2. INTERSECTION mode: A ray must collide with the surface of the
 * intersection with all atomic (non-glue) components.
 * FIXME: caller is responsible for making sure C is big enough.
 */
int glue::
detectCollisions( const ray *R, collision C[] ) const
{
	collision *D, *all, **ptr, *ap;
	int k, m, n, num_atoms;
	typedef int (*cmpfn)(const void *, const void *);

	switch(combine_mode) {
	case UNION:
		k = 0;
		for (int i = 0; i < num_components; ++i) {
			D = new collision[2*component[i]->size()];
			m = component[i]->detectCollisions(R, D);
			for (int j = 0; j < m; ++j)
				C[k++] = D[j];
			delete D;
			qsort(C, k, sizeof(collision), (cmpfn)collisionCmp);
		}

		return k;
	case INTERSECTION:
		num_atoms = 0;
		for (int i = 0; i < num_components; ++i)
			num_atoms += component[i]->size();
		all = new collision[2*num_atoms];
		ptr = new (collision *)[2*num_atoms];

		k = 0;
		m = 0;
		ap = all;
		for (int i = 0; i < num_components; ++i) {
			/* keep track of how many regions we are already inside */
			/* TODO: implement inverse figures */
			if (component[i]->inside(R->R0))
				++m;

			n = component[i]->detectCollisions(R, ap);
			if (n == 0) continue;
			
			/* build an array of pointers to components, to make sorting
			 * easier.  Ignore collisions behind the camera. */
			for (int j = 0; j < n; ++j)
				if (ap->rayPosition > 0)
					ptr[k++] = ap++;
				else
					ap++;
		}

		/* sort so that "earlier" collisions are first */
		qsort(ptr, k, sizeof(collision *), (cmpfn)collisionPtrCmp);

		for (int i = 0; i < k; ++i)
			if ((m += ptr[i]->increment) == num_atoms) {
				ptr[i]->computeNormal();
				C[0] = *ptr[i];
				delete ptr;
				delete all;
				return 1;
			}

		delete ptr;
		delete all;
		break;
	}

	return 0;
}

/**
 * Set the color of every component of this glue.
 */
void glue::
setColor( unsigned char r, unsigned char g, unsigned char b )
{
	for (int i = 0; i < num_components; ++i)
		component[i]->setColor(r, g, b);
}

/** 
 * Translate this glue.  This works by translating every component.
 */
void glue::
translate( const vector4 &c )
{
	for (int i = 0; i < num_components; ++i)
		component[i]->translate(c);
}
