/*
 *  This file is part of Dascant.
 *
 *  Dascant is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Dascant is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Dascant; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
 *  USA
 */

/**
 * The GTK+-based GUI.
 */
#include <gui.h>
#include <config.h>
#include <dascant.h>

#include <stdio.h>
#include <stdlib.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h>
#include <math.h>

extern int siz;
extern bool anim;
extern entity *scene;

static GtkWidget *drawing_area;

/**
 * Handle exposure events.
 */
static int
exposure( GtkWidget *da, GdkEventExpose *ev, unsigned char *pix )
{
	gdk_draw_rgb_image(da->window, da->style->black_gc,
		0, 0, siz, siz, GDK_RGB_DITHER_NONE, pix, 3*siz);

	return 0;
}

/**
 * Handle mouse button presses.  FIXME: pretty much doesn't work at all.
 */
static int
button_press( GtkWidget *da, GdkEventButton *ev, camera *C )
{
	vector4 P = C->position;

	switch(ev->button) {
	case 1: // change xy coords.
		P.elem[0] = 0.05 * (ev->x - C->width  / 2);
		P.elem[1] = 0.05 * (ev->y - C->height / 2);
		break;
	case 2:
		P.elem[2] = 0.05 * (ev->y - C->height / 2);
		P.elem[3] = 0.05 * (ev->x - C->width  / 2);
		break;
	}

	C->enqueuePosition(&P);

	return 0;
}

/**
 * Handle a keyboard event.  This currently covers camera movement and
 * rotation, and exiting.
 */
static int
key_press( GtkWidget *da, GdkEventKey *ev, camera *C )
{
	vector4 v;
	frame *A = C->view;

	switch(ev->keyval) {
	/* miscellaneous */
	case KEY_QUIT:   exit(0); break;
	case KEY_RETURN:
		A->origin(0, 0, 10, 0);
		for (int i = 0; i < 4; ++i)
			for (int j = 0; j < 4; ++j)
				A->local[i][j] = i == j;
		break;
	/* translation */
	case KEY_RIGHT:    v = A->local[0] * rcfile::move_stepsize; break;
	case KEY_LEFT:     v = A->local[0] * -rcfile::move_stepsize; break;
	case KEY_UP:       v = A->local[1] * rcfile::move_stepsize; break;
	case KEY_DOWN:     v = A->local[1] * -rcfile::move_stepsize; break;
	case KEY_FORWARD:  v = A->local[2] * -rcfile::move_stepsize; break;
	case KEY_BACKWARD: v = A->local[2] * rcfile::move_stepsize; break;
	case KEY_ANA:      v = A->local[3] * rcfile::move_stepsize; break;
	case KEY_KATA:     v = A->local[3] * -rcfile::move_stepsize; break;
	/* rotation */
	case KEY_YAW_LEFT:		A->rotate(1, 3);	break;
	case KEY_YAW_RIGHT:		A->rotate(1, 3, 1);	break;
	case KEY_PITCH_UP:		A->rotate(0, 3, 1);	break;
	case KEY_PITCH_DOWN:	A->rotate(0, 3);	break;
	case KEY_ROLL_RIGHT:	A->rotate(2, 3, 1);	break;
	case KEY_ROLL_LEFT:		A->rotate(2, 3);	break;
	case KEY_SPILL_XY_ANA:	A->rotate(0, 1);	break;
	case KEY_SPILL_XY_KATA:	A->rotate(0, 1, 1);	break;
	case KEY_SPILL_XZ_ANA:	A->rotate(0, 2);	break;
	case KEY_SPILL_XZ_KATA:	A->rotate(0, 2, 1);	break;
	case KEY_SPILL_YZ_ANA:	A->rotate(1, 2);	break;
	case KEY_SPILL_YZ_KATA:	A->rotate(1, 2, 1);	break;
	}

	A->origin += v;
	C->compute_rays();
	if (!anim) {
		C->snapshot(scene);
		gui_update();
	}

	return 0;
}

/**
 * Reset button on the parameter dialog clicked.
 */
static int
optReset_clicked( GtkWidget *w, GtkEntry *entry[] )
{
	char buf[256];

	sprintf(buf, "%.4f", rcfile::move_stepsize);
	gtk_entry_set_text(entry[0], buf);

	sprintf(buf, "%.1f", rcfile::rotate_stepsize * 180 / M_PI);
	gtk_entry_set_text(entry[1], buf);

	for (int i = 0; i < 4; ++i) {
		sprintf(buf, "%.2f", rcfile::lightSource[i]);
		gtk_entry_set_text(entry[i+2], buf);
	}

	return 0;
}

/**
 * Update button on the parameters dialog clicked.
 */
static int
optUpdate_clicked( GtkWidget *w, GtkEntry *entry[] )
{
	rcfile::move_stepsize = strtod(gtk_entry_get_text(entry[0]), NULL);

	double x = strtod(gtk_entry_get_text(entry[1]), NULL);
	rcfile::rotate_stepsize = x * M_PI / 180;

	for (int i = 0; i < 4; ++i)
		rcfile::lightSource[i]
			= strtod(gtk_entry_get_text(entry[i+2]), NULL);
	rcfile::lightSource.normalize();
	optReset_clicked(w, entry);

	gui_update();

	return 1;
}

/** 
 * Load button in the parameter dialog clicked.
 */
int
optLoad_clicked( GtkWidget *w, GtkEntry *entry[] )
{
	rcfile::load();
	optReset_clicked(NULL, entry);

	return 0;
}

/** 
 * Save button in the parameter dialog clicked
 */
int
optSave_clicked( GtkWidget *w, GtkEntry *entry[] )
{
	optUpdate_clicked(NULL, entry);
	rcfile::save();

	gui_update();

	return 0;
}

/**
 * Build the GUI.
 */
void
gui_init( camera *C )
{
	GtkWidget *win;
	GtkWidget *optWin, *optTable, *optReset, *optUpdate, *optSave, *optLoad;
	static GtkWidget *optEntry[6];

	/* make display window */
	win = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	g_signal_connect(G_OBJECT(win), "delete-event",
		G_CALLBACK(exit), 0);

	drawing_area = gtk_drawing_area_new();
	gtk_drawing_area_size(GTK_DRAWING_AREA(drawing_area), siz, siz);
	g_signal_connect(G_OBJECT(drawing_area), "expose-event",
		G_CALLBACK(exposure), C->pixel);
	g_signal_connect(G_OBJECT(drawing_area), "button-press-event",
		G_CALLBACK(button_press), C);
	g_signal_connect(G_OBJECT(drawing_area), "key-press-event",
		G_CALLBACK(key_press), C);
	gtk_widget_set_events(drawing_area, GDK_EXPOSURE_MASK
	                          |GDK_BUTTON_PRESS_MASK
							  |GDK_KEY_PRESS_MASK);
	GTK_WIDGET_SET_FLAGS(drawing_area, GTK_CAN_FOCUS);
	gtk_container_add(GTK_CONTAINER(win), drawing_area);

	gtk_widget_show_all(win);

	/* make options window */
	optWin = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	g_signal_connect(G_OBJECT(optWin), "delete-event",
		G_CALLBACK(exit), 0);

	optTable = gtk_table_new(4, 5, 0);
	gtk_container_add(GTK_CONTAINER(optWin), optTable);

	gtk_table_attach(GTK_TABLE(optTable), gtk_label_new("Move stepsize"),
		0, 1, 0, 1, GTK_SHRINK, GTK_SHRINK, 2, 2);
	gtk_table_attach(GTK_TABLE(optTable), gtk_label_new("Rotate stepsize"),
		0, 1, 1, 2, GTK_SHRINK, GTK_SHRINK, 2, 2);
	gtk_table_attach(GTK_TABLE(optTable), gtk_label_new("Light source"),
		0, 1, 2, 3, GTK_SHRINK, GTK_SHRINK, 2, 2);

	char buf[256];

	optEntry[0] = gtk_entry_new();
	gtk_widget_set_size_request(optEntry[0], 4*(ENTRY_WIDTH+4), -1);
	sprintf(buf, "%.4f", rcfile::move_stepsize);
	gtk_entry_set_text(GTK_ENTRY(optEntry[0]), buf);
	gtk_table_attach(GTK_TABLE(optTable), optEntry[0],
		1, 5, 0, 1, GTK_SHRINK, GTK_SHRINK, 2, 2);

	optEntry[1] = gtk_entry_new();
	gtk_widget_set_size_request(optEntry[1], 4*(ENTRY_WIDTH+4), -1);
	sprintf(buf, "%.1f", rcfile::rotate_stepsize * 180 / M_PI);
	gtk_entry_set_text(GTK_ENTRY(optEntry[1]), buf);
	gtk_table_attach(GTK_TABLE(optTable), optEntry[1],
		1, 5, 1, 2, GTK_SHRINK, GTK_SHRINK, 2, 2);

	for (int i = 0; i < 4; ++i) {
		optEntry[i+2] = gtk_entry_new();
		gtk_widget_set_size_request(optEntry[i+2], ENTRY_WIDTH, -1);
		sprintf(buf, "%.2f", rcfile::lightSource[i]);
		gtk_entry_set_text(GTK_ENTRY(optEntry[i+2]), buf);
		gtk_table_attach(GTK_TABLE(optTable), optEntry[i+2],
			i+1, i+2, 2, 3, GTK_SHRINK, GTK_SHRINK, 2, 2);
	}

	optReset = gtk_button_new_with_label("Reset");
	g_signal_connect(G_OBJECT(optReset), "clicked",
		G_CALLBACK(optReset_clicked), optEntry);
	gtk_table_attach(GTK_TABLE(optTable), optReset,
		1, 2, 3, 4, GTK_SHRINK, GTK_SHRINK, 2, 2);

	optUpdate = gtk_button_new_with_label("Update");
	g_signal_connect(G_OBJECT(optUpdate), "clicked",
		G_CALLBACK(optUpdate_clicked), optEntry);
	gtk_table_attach(GTK_TABLE(optTable), optUpdate,
		2, 3, 3, 4, GTK_FILL, GTK_FILL, 2, 2);
	GTK_WIDGET_SET_FLAGS(optUpdate, GTK_CAN_DEFAULT);
	gtk_window_set_default(GTK_WINDOW(optWin), optUpdate);

	optLoad = gtk_button_new_with_label("Load");
	g_signal_connect(G_OBJECT(optUpdate), "clicked",
		G_CALLBACK(optLoad_clicked), optEntry);
	gtk_table_attach(GTK_TABLE(optTable), optLoad,
		3, 4, 3, 4, GTK_FILL, GTK_FILL, 2, 2);
	GTK_WIDGET_SET_FLAGS(optLoad, GTK_CAN_DEFAULT);
	gtk_window_set_default(GTK_WINDOW(optWin), optLoad);

	optSave = gtk_button_new_with_label("Save");
	g_signal_connect(G_OBJECT(optUpdate), "clicked",
		G_CALLBACK(optSave_clicked), optEntry);
	gtk_table_attach(GTK_TABLE(optTable), optSave,
		4, 5, 3, 4, GTK_FILL, GTK_FILL, 2, 2);
	GTK_WIDGET_SET_FLAGS(optSave, GTK_CAN_DEFAULT);
	gtk_window_set_default(GTK_WINDOW(optWin), optSave);

	gtk_widget_show_all(optWin);
}

/**
 * update the image by invalidating the drawing area.
 */
void
gui_update( void )
{
	static GdkRectangle rect = { 0, 0, siz, siz };

	gdk_window_invalidate_rect(drawing_area->window, &rect, 0);
}
