/*
 *  This file is part of Dascant.
 *
 *  Dascant is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Dascant is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Dascant; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
 *  USA
 */

#ifndef _DASCANT_ROTATION4_H
#define _DASCANT_ROTATION4_H

/**
 * The rotation4 class.  This class represents a rotation in tetraspace.
 *
 * note: a "canonical" rotation is one that is defined as a single rotation
 * matrix R.  The only such rotations possible using this class are those
 * around the coordinate planes.  A noncanonical rotation will be of the
 * form  M^(-1) R_xy M.  However, the class itself will generate other
 * canonical rotations, one column at a time; namely, the M matrix itself.
 * It must be considered canonical so that it does not need its own M
 * matrix. */

#include <transform4.h>

class rotation4 : public transform4 {
private:
	/**
	 * the angle of rotation.
	 */
	double theta;

	/**
	 * the rule according to which the signs of sine terms are assigned.
	 */
	static const int signRule[][4];

	bool canonical; /* are we rotating around one of the
	                            * coordinate planes? */

	/**
	 * The axes of the plane around which the rotation the rotation is to be
	 * performed.
	 */
	int axes, inFlumeAxis[4];

	/**
	 * Extra information for noncanonical rotations.
	 */
	rotation4 *M, *Minv, *MiRxyM;
	void construct(int a1, int a2);
public:
	/**
	 * axis specifiers, for generating canonical rotations.
	 */
	static const int xAxis=1, yAxis=2, zAxis=4, wAxis=8;
	rotation4(void);
	rotation4(vector4 &U, vector4 &V);
	rotation4(int a1, int a2);

	void setAngle(double t);
	double getAngle(void);

	vector4 applyTo(double t, vector4 &P);
	transform4 applyTo(double t, transform4 &T);
	rotation4 applyTo(double t, rotation4 &T);
	vector4 applyTo(vector4 &);
	rotation4 applyTo(rotation4 &);
	transform4 applyTo(transform4 &);

	static int toAxis(const char *s);
	rotation4 inverse(void);
	rotation4 transpose(void);
	rotation4 &operator=(transform4 &);
};

#endif
