/*
 *  This file is part of Dascant.
 *
 *  Dascant is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Dascant is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Dascant; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
 *  USA
 */

/**
 * A quadric surface (or solid).  Defined by F(x,y,z,w) = 0 (or <= 0 for
 * solids); F must be a polynomial of degree no higher than two,
 */
#include <quadric.h>
#include <collision.h>
#include <math.h>
#include <stdio.h>

/** 
 * Default constructor.  Creates a quadric surfaces with zero coefficients.
 */
quadric::
quadric(void)
	: entity()
{

	int x, y, z, w;

	for (x = 0; x <= 2; ++x)
		for (y = 0; y <= 2-x; ++y)
			for (z = 0; z <= 2-x-y; ++z)
				for (w = 0; w <= 2-x-y-z; ++w)
					coeff[x][y][z][w] = 0.0;

	normalOrientation = 1;
}

#if 0
/**
 * Return the coefficient of the <tt>trm</tt>th term.
 * @param trm a Term object used to index the coefficient array.
 * @return the coefficient of the index term.
 */
double quadric::
getCoefficient( term trm )
{
	if (trm.degree() > 2)
		throw new GeometryException("For quadric surfaces, "
		 + "no term can have degree greater than 2, but this "
		 + "term is " + trm);
	return coeff[trm.powX][trm.powY][trm.powZ][trm.powW];
}

/**
 * Set the coefficient of the <tt>trm</tt>th term to <tt>c</tt>.
 * @param c the new coefficient.
 * @param trm an index to the desired term.
 */
void quadric::
setCoefficient( double c, Term trm )
	throws GeometryException
{
	if (trm.degree() > 2)
		throw new GeometryException("For quadric surfaces, "
		 + "no term can have degree greater than 2, but this "
		 + "term is " + trm);
	coeff[trm.powX][trm.powY][trm.powZ][trm.powW] = c;
}

/**
 * Add term <tt>T</tt> to this quadric.
 * @param T the term to add.
 */
void quadric::
add( Term T ) 
	throws GeometryException
{
	if (T.degree() > 2)
		throw new GeometryException("For quadric surfaces, "
		 + "no term can have degree greater than 2, but this "
		 + "term is " + T);
	coeff[T.powX][T.powY][T.powZ][T.powW] += T.coeff;
}
#endif

/**
 * Detect collisions.  Do this by substituting the parametrized form of
 * the ray into the equation for the surface to derive a quadric
 * equation in <i>t</i>, the parameter for the ray.  Solve this
 * equation, and use the solutions to determine all collisions with the
 * surface.
 * @param R a ray to check for collisions with.
 * @return a collision object, or <tt>null</tt> if no collisions occur.
 */
int quadric::
detectCollisions( const ray *R, collision *C ) const
{
	double a, b, c, desc;
	int i, n;

	// find quadratic equation  a t^2 + b t + c = 0  by substituting
	// parametric form of ray into this quadric.  ray conveniently
	// provides us with the coefficients of t^2, t, and 1 for each
	// possible product of x, y, z, and w, so we only need to multiply
	// those products by the coefficients of the terms in the equation
	// of the surface.
	a = b = c = 0;
	for (i = 0; i < 15; ++i) {
		a += coeff[term[i][0]][term[i][1]][term[i][2]][term[i][3]]
		   * R->ttab[i][2];
		b += coeff[term[i][0]][term[i][1]][term[i][2]][term[i][3]]
		   * R->ttab[i][1];
		c += coeff[term[i][0]][term[i][1]][term[i][2]][term[i][3]]
		   * R->ttab[i][0];
	}

	// there has to be at least one term in the quadtratic equation with
	// a t in it.
	if (fabs(a) < 1e-16 && fabs(b) < 1e-16)
		return 0;

	desc = b*b - 4*a*c;
	if (desc < 0) return 0;

	double sqrtDesc = sqrt(desc);
	if (sqrtDesc < 1e-16) {
		n = 1;

		C[0](R, -b/(2*a), this);
		C[0].surface = this;
		C[0].increment = 1;
	} else {
		n = 2;

		C[0](R, (-b - sqrtDesc) / (2*a), this);
		C[0].surface = this;

		C[1](R, (-b + sqrtDesc) / (2*a), this);
		C[1].surface = this;

		/* TODO: take normalOrientation into account */
		if (C[0].rayPosition * C[1].rayPosition < 0) {
		         /* start inside */
			C[1].increment = -1;
			C[0].increment = 0;
		} else { /* start outside */
			if (C[0].rayPosition > 0) {
				C[0].increment = +1;
				C[1].increment = -1;
			} else {
				C[0].increment = 0;
				C[1].increment = 0;
			}
		}
	}
	/* TODO: Figure out outgoing ray, insideness. */

	return n;
}

/**
 * Add to this quadric another.  Simply adds corresponding coefficients.
 * @param Q the other quadric to add to this one.
 */
void quadric::
add( quadric &Q )
{
	int i, j, k, l;

	for (i = 0; i <= 2; ++i)
		for (j = 0; j <= 2-i; ++j)
			for (k = 0; k <= 2-i-j; ++k)
				for (l = 0; l <= 2-i-j-k; ++l)
					coeff[i][j][k][l] += Q.coeff[i][j][k][l];
}

/**
 * Return, as a string, the equation defining this quadric surface.
 * @return the equation as a string.
 */
char *quadric::
toString( char *s=0x0 ) const
{
	int i, j, k, l;
	double c;
	static char buf[256], *cp;
	bool first=true;

	if (!s) s = buf;
	cp = s;

	for (i = 2; i >= 0; --i)
		for (j = 2-i; j >= 0; --j)
			for (k = 2-i-j; k >= 0; --k)
				for (l = 2-i-j-k; l >= 0; --l) {
					if (i+j+k+l == 0) continue;

					c = coeff[i][j][k][l];
					if (fabs(c) < 1e-16)
						continue;

					if (c < 0)                { *cp++ = '-'; *cp++ = ' '; }
					else if (c > 0 && !first) { *cp++ = '+'; *cp++ = ' '; }
					first = false;

					c = fabs(c);
					if (fabs(c-1) > 1e-16)
						cp += sprintf(cp, "%g", c);

					if (i > 0) cp += sprintf(cp, "x%s ", i>1 ? "^2" : "");
					if (j > 0) cp += sprintf(cp, "y%s ", j>1 ? "^2" : "");
					if (k > 0) cp += sprintf(cp, "z%s ", k>1 ? "^2" : "");
					if (l > 0) cp += sprintf(cp, "w%s ", l>1 ? "^2" : "");
				}

	sprintf(cp, "%s %g",
	     (normalOrientation < 0 ? ">=" : "<="),
	     -coeff[0][0][0][0]);

	return s;
}

const int quadric::term[][4] = {
	{ 2, 0, 0, 0 }, // square terms
	{ 0, 2, 0, 0 },
	{ 0, 0, 2, 0 },
	{ 0, 0, 0, 2 },
	{ 1, 1, 0, 0 }, // cross terms
	{ 1, 0, 1, 0 },
	{ 1, 0, 0, 1 },
	{ 0, 1, 1, 0 },
	{ 0, 1, 0, 1 },
	{ 0, 0, 1, 1 },
	{ 1, 0, 0, 0 }, // linear terms
	{ 0, 1, 0, 0 },
	{ 0, 0, 1, 0 },
	{ 0, 0, 0, 1 },
	{ 0, 0, 0, 0 }, // constant term
};

const int quadric::degree[] = {
	2, 2, 2, 2,
	2, 2, 2, 2, 2, 2,
	1, 1, 1, 1,
	0
};

/**
 * Swap the orientation of this quadric (F --> -F).
 */
void quadric::
invert( void )
{
	int i, j, k, l;

	for (i = 0; i <= 2; ++i)
		for (j = 0; j <= 2-i; ++j)
			for (k = 0; k <= 2-i-j; ++k)
				for (l = 0; l <= 2-i-j-k; ++l)
					coeff[i][j][k][l] *= -1;
}

inline double toPowOf( double x, int y )
{
	switch(y) {
	case 0: return 1;
	case 1: return x;
	case 2: return x*x;
	}
	return 0;
}

/**
 * Determine whether the given point is inside this quadric.
 */
bool quadric::
inside( const vector4 &V ) const
{
	double s=0, t;

	for (int i = 0; i < 15; ++i) {
		t = coeff[term[i][0]][term[i][1]][term[i][2]][term[i][3]];
		for (int m = 0; m < 4; ++m)
			//t *= toPowOf(V.elem[m], term[i][m]);
			t *= pow(V.elem[m], term[i][m]);
		s += t;
	}

	return s < 0;
}

/** 
 * Swap the orientation of this quadric.  Performed by setting some
 * parameter.  Why do I have this and "invert"?  I don't know.
 */
void quadric::
setComplement( bool c )
{
	normalOrientation = c ? -1 : +1;
}

/**
 * Translate this quadric.
 * TODO: consider cross-terms.
 * TODO: insist translation occurs *after* rotations
 */
void quadric::
translate( const vector4 &c )
{
	coeff[1][0][0][0] = -2*coeff[2][0][0][0] * c.elem[0];
	coeff[0][1][0][0] = -2*coeff[0][2][0][0] * c.elem[1];
	coeff[0][0][1][0] = -2*coeff[0][0][2][0] * c.elem[2];
	coeff[0][0][0][1] = -2*coeff[0][0][0][2] * c.elem[3];

	//coeff[0][0][0][0] += c.dot(c);
	coeff[0][0][0][0] += coeff[2][0][0][0]*c.elem[0]*c.elem[0]
	                  +  coeff[0][2][0][0]*c.elem[1]*c.elem[1]
	                  +  coeff[0][0][2][0]*c.elem[2]*c.elem[2]
	                  +  coeff[0][0][0][2]*c.elem[3]*c.elem[3];
}

/** 
 * Rotate this quadric.
 * FIXME: not actually implented yet.
 */
void quadric::
rotate( axis u, axis v, double t )
{
	double A[2], C[4][4], D[2];
	double c=cos(t), s=sin(t), c2=c*c, s2=s*s;

	/* get square terms' coefficients */
	A[0] = coeff[2*(u==0)][2*(u==1)][2*(u==2)][2*(u==3)];
	A[1] = coeff[2*(v==0)][2*(v==1)][2*(v==2)][2*(v==3)];

	/* get cross terms' coefficients */
	C[u][v] = coeff[u==0||v==0][u==1||v==1][u==2||v==2][u==3||v==3];
	for (int i = 0; i < 4; ++i) {
		if (u != i)
			C[u][i] = coeff[u==0||i==0][u==1||i==1][u==2||i==2][u==3||i==3];
		if (v != i)
			C[v][i] = coeff[i==0||v==0][i==1||v==1][i==2||v==2][i==3||v==3];
	}

	/* get linear terms' coefficients */
	D[0] = coeff[u==0][u==1][u==2][u==3];
	D[1] = coeff[v==0][v==1][v==2][v==3];

	/* apply changes to square terms' coefficients */
	coeff[2*(u==0)][2*(u==1)][2*(u==2)][2*(u==3)] = A[0] * c2;
	coeff[2*(v==0)][2*(v==1)][2*(v==2)][2*(v==3)] = A[1] * c2;

	/* apply changes to cross terms' coefficients */
	coeff[u==0||v==0][u==1||v==1][u==2||v==2][u==3||v==3]
		= 2*(A[1]-A[0])*c*s + C[u][v]*(1-2*s2);
	for (int k = 0; k < 4; ++k) {
		if (u != k)
			coeff[u==0||k==0][u==1||k==1][u==2||k==2][u==3||k==3]
				=  C[u][k]*c + C[v][k]*s;
		if (v != k)
			coeff[v==0||k==0][v==1||k==1][v==2||k==2][v==3||k==3]
				= -C[u][k]*s + C[v][k]*c;
	}

	/* apply changes to cross terms' coefficients */
	coeff[u==0][u==1][u==2][u==3] =  D[0]*c + D[1]*s;
	coeff[v==0][v==1][v==2][v==3] = -D[0]*s + D[1]*c;
}
