/*
 *  This file is part of Dascant.
 *
 *  Dascant is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Dascant is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Dascant; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
 *  USA
 */

/**
 * The ray class.  For raytracing.
 */

#include <stdio.h>
#include <ray.h>
#include <camera.h>
#include <quadric.h>

/**
 * Class constructor.  Generate a ray from the origin through the point
 * (<tt>x</tt>,<tt>y</tt>) through the (negative) image plane.
 * @param C the Camera receiving this ray.
 * @param X x-coordinate of point on image plane.
 * @param Y y-coordinate of point on image plane.
 */
ray::
ray( camera *C, double X, double Y )
	: vector4(X, Y, -C->depth)
{
	double a, b, c;
	double  t[2], // coeffs of first parametric term
	       tt[2]; // coeffs of second parametric term
	int i, m;
	bool gott;

	/* want ray to have unit length */
	//normalize();

	/* want ray to point in the right direction */
	vector4 v = C->view->toAbsolute(*this);
	for (i = 0; i < 4; ++i)
		elem[i] = v.elem[i];

	R0 = C->view->origin;

	/* build the array of coefficients derived from the parametrized
	 * form of this ray. */
	for (i = 0; i < 15; ++i) {
		gott = false;

		// figure out what linear terms (in t) will be
		// multiplied together.
		for (m = 0; m < 4; ++m) {
			switch(quadric::term[i][m]) {
			case 2:
				 t[1] = elem[m];
				 t[0] = R0.elem[m];
				tt[1] = elem[m];
				tt[0] = R0.elem[m];
				m = 4; // we can break out of the loop now
				break;
			case 1:
				if (gott) {
					tt[1] = elem[m];
					tt[0] = R0.elem[m];
					break;
				}
				else {
					t[1] = elem[m];
					t[0] = R0.elem[m];
					gott = true;
				}
				break;
			}
		}

		// actually multiply them together
		switch(quadric::degree[i]) {
		case 0:
			a = 0;
			b = 0;
			c = 1;
			break;
		case 1:
			a = 0;
			b = t[1];
			c = t[0];
			break;
		case 2:
			a = t[1]*tt[1];
			b = t[0]*tt[1] + t[1]*tt[0];
			c = t[0]*tt[0];
			break;
		default:
			/* FIXME: i probably shouldn't just ignore this
			 * possibility */
			return;
		}

		// store the results in the table
		ttab[i][0] = c;
		ttab[i][1] = b;
		ttab[i][2] = a;
	} /* for(...) */
}

/**
 * Determine what point in space is distance t away from the camera center
 * along this ray.
 */
vector4 &ray::
evalAt( double t ) const
{
	static vector4 U;

	for (int i = 0; i < 4; ++i)
		U.elem[i] = elem[i]*t + R0.elem[i];

	return U;
}

/** 
 * Return a string representation of this ray.
 */
char *ray::
toString( char *str )
{
	static char buf[3][256];

	if (!str) str = buf[2];

	R0.toString(buf[0]);
	((vector4 *)this)->toString(buf[1]);
	sprintf(str, "[ %s + %s t ]", buf[0], buf[1]);

	return str;
}

/**
 * Dummy constructor; should not be used by users.
 */
ray::
ray()
{
} 

#if 0
ray ray::
copy( camera C, int X, int Y )
{
	ray R;

	R.elem[0] = X;
	R.elem[1] = Y;
	R.elem[2] = -C.depth;
	R.elem[3] = 0;

	R.normalize();
	R.R0 = C.position;

	R.ttab = new double[15][3];
	for (int i = 0; i < 15; ++i)
		for (int j = 0; j < 3; ++j)
			R.ttab[i][j] = ttab[i][j];

	return R;
}
#endif

/**
 * Reinitialize this ray.
 */
ray &ray::operator
()( camera *C, double x, double y )
{
	ray r(C, x, y);
	*this = r;
	return *this;
}
