/*
 *  This file is part of Dascant.
 *
 *  Dascant is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Dascant is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Dascant; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
 *  USA
 */

/**
 * Rotation class.  A lot of this functionality (esp. rotation around
 * arbitrary planes) is deprecated and broken anyway.
 */
#include <rotation4.h>
#include <vector4.h>
#include <math.h>
#include <string.h>

/**
 * Class constructor.  The default constructor generates a 90-degree
 * rotation around the <i>x-y</i> plane.
 */
rotation4::
rotation4(void)
{
/* FIXME */
#if 0
	try { construct(xAxis, yAxis); }
	catch (GeometryException e) {
		System.err.println(e.getMessage());
		System.exit(1);
	}
#else
	construct(xAxis, wAxis);
#endif
	canonical = true;
}

/**
 * Class constructor to create a 90-degree non-canonical rotation (i.e.
 * one not oriented around one of the coordinate planes).
 * @param U one vector in the axis plane; should be unit length.
 * @param V another vector orthogonal to to <tt>U</tt>; should also be
 * unit length.
 */
rotation4::
rotation4( vector4 &u, vector4 &v )
{
/* FIXME */
#if 0
	/* check for orthogonality of U and V */
	if (U.dot(V) > 1e-16)
		throw new GeometryException("U and V should be orthogonal.");
#endif

	/* check for normality */
	vector4 U=u, V=v;
	U.normalize();
	V.normalize();

	canonical = false;

	construct(xAxis, yAxis);

	vector4 Zi, Wi;
	double z, w;

	z = - (U.elem[0]*V.elem[3] - V.elem[0]*U.elem[3])
	    / (U.elem[2]*V.elem[3] - V.elem[2]*U.elem[3]);
	w = -1/U.elem[3] * (V.elem[1] + V.elem[2]*z);
	Zi(0, 1, z, w);
	Wi = U.cross(V, Zi);

	/* note: the M and Minv matrices need to be considered canonical */
	M = new rotation4;
	M->setColumn(0, U);
	M->setColumn(1, V);
	M->setColumn(2, Zi);
	M->setColumn(3, Wi);

	Minv = new rotation4;
	*M = M->inverse();
}

/**
 * Class constructor to generate a 90-degree canonical transformation
 * around one of the coordinate planes.
 * @param a1 one of the coordinate axes; for example,
 * <tt>rotation4.xAxis</tt>.
 * @param a2 the other coordinate axes, to define a plane.
 */
rotation4::
rotation4( int a1, int a2 )
{
	construct(a1, a2);
	canonical = true;
}

/**
 * Constructor helper function for canonical rotations.
 */
void rotation4::
construct( int a1, int a2 )
{
	int i, j=0;

	canonical = true;

	axes = a1 | a2;

	for (i = 0; i < 4; ++i) {
		if ((axes & (1<<i)) != 0)
			matrix[i][i] = 1;
		else
			inFlumeAxis[j++] = i;
	}
#if 0
	if (j != 2)
		throw new GeometryException("Need exactly two axis "
			+ "(one plane) to rotate around.");
#endif

	matrix[inFlumeAxis[0]][inFlumeAxis[1]] = 
		signRule[inFlumeAxis[0]][inFlumeAxis[1]];
	matrix[inFlumeAxis[1]][inFlumeAxis[0]] = 
		signRule[inFlumeAxis[1]][inFlumeAxis[0]];

	matrix[inFlumeAxis[0]][inFlumeAxis[0]] = 0;
	matrix[inFlumeAxis[1]][inFlumeAxis[1]] = 0;
}

/**
 * Change the angle through which vectors will be rotated.
 * @param t the new angle to rotate through.
 */
void rotation4::
setAngle( double t )
{
	theta = t;
	double st = sin(theta), ct = cos(theta);

	if (fabs(ct) < 1e-16) ct = 0;
	if (fabs(st) < 1e-16) st = 0;

	matrix[inFlumeAxis[0]][inFlumeAxis[1]] = 
		signRule[inFlumeAxis[0]][inFlumeAxis[1]] * st;
	matrix[inFlumeAxis[1]][inFlumeAxis[0]] = 
		signRule[inFlumeAxis[1]][inFlumeAxis[0]] * st;

	matrix[inFlumeAxis[0]][inFlumeAxis[0]] = ct;
	matrix[inFlumeAxis[1]][inFlumeAxis[1]] = ct;

	if (canonical) return;

	/* if we're not dealing with one of the "canonical" rotations (i.e.
	 * those around coordinate planes) then we need to update the matrix
	 * that's actually going to do the rotation. */
/* FUCK XXX
	MiRxyM = new rotation4;
	transform4 R = applyTo(*M), *S = Minv;
	*MiRxyM = *(rotation4 *)(S->applyTo(R));
 */
}

/**
 * Return the angle of rotation.
 * @return the angle of rotation.
 */
double rotation4::
getAngle(void)
{
	return theta;
}

/**
 * Apply this rotation to a vector, allowing specification of angle.
 * @param t the angle to rotate through.
 * @param P the vector to rotate.
 * @return a vector4 containing the image of <tt>P</tt> under rotation
 * through <tt>t</tt>.
 */
vector4 rotation4::
applyTo( double t, vector4 &P )
{
	setAngle(t);
	if (canonical)
		return applyTo(P);
	else
		return MiRxyM->applyTo(P);
} 

/**
 * Compose this rotation with another transformation, allowing
 * specification of angle.
 * @param t the angle to rotate through.
 * @param T the transformation to compose with.
 * @return a transform4 representing the composition of this rotation
 * with <tt>T</tt>.
 */
transform4 rotation4::
applyTo( double t, transform4 &T )
{
	setAngle(t);
	if (canonical)
		return applyTo(T);
	else
		return MiRxyM->applyTo(T);
}

/**
 * Return an axis specifier given by a string.
 * @param s a String containing only one of the letters x, y, z, or w,
 * in upper- or lower-case.
 * @return an axis specifier (such as <tt>rotation4.xAxis</tt>).
 */
int rotation4::
toAxis( const char *s )
{
	if (!strcasecmp(s, "x")) return xAxis;
	if (!strcasecmp(s, "y")) return yAxis;
	if (!strcasecmp(s, "z")) return zAxis;
	if (!strcasecmp(s, "w")) return wAxis;
	return -1;
/* FIXME */
#if 0
	throw new GeometryException("Invalid axis specification.");
#endif
}

/**
 * Return the inverse of this transformation.
 * @return a new rotation4 representing the inverse of this rotation.
 */
rotation4 rotation4::
inverse(void)
{
	return transpose();
}

/**
 * Return the transposition of this transformation.
 * @return a new rotation4 representing the transformation formed by
 * the transposition of the matrix of this rotation.
 */
rotation4 rotation4::
transpose(void)
{
	rotation4 T;
	int i;

	for (i = 0; i < 4; ++i) {
		vector4 V = getColumn(i);
		T.setRow(i, V);
	}

	return T;
}

/* FIXME: wtf do I have to do this? */
vector4 rotation4::
applyTo( vector4 &P )
{
	transform4 *T = (transform4 *)this;
	return T->applyTo(P);
}

transform4 rotation4::
applyTo( transform4 &P )
{
	transform4 *T = (transform4 *)this;
	return T->applyTo(P);
}

rotation4 rotation4::
applyTo( rotation4 &R )
{
	transform4 *T = (transform4 *)this;
	rotation4 S;

	*T = T->applyTo(R);
	memcpy(S.matrix, T->matrix, 4*4*sizeof(double));
	S.theta = theta;
	return S;
}

const int rotation4::
signRule[][4] = {
	{  1, -1,  1, -1 },
	{  1,  1, -1,  1 },
	{ -1,  1,  1, -1 },
	{  1, -1,  1,  1 }
};

rotation4 &rotation4::
operator=( transform4 &T )
{
	memcpy(matrix, T.matrix, sizeof(T.matrix));
	theta = M_PI/2;
	return *this;
}
