/*
 *  This file is part of Dascant.
 *
 *  Dascant is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Dascant is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Dascant; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
 *  USA
 */

/**
 * The tesseract class.  Represents a four-dimensional hyperbox.
 */

#include <stdlib.h>
#include <stdio.h>

#include <tesseract.h>
#include <ray.h>
#include <vector4.h>
#include <collision.h>
#include <quadric.h>
#include <cucumber.h>

/**
 * Class constructor allowing specification of all the dimensions of the
 * tesseract.
 * @param s a vector containing the dimensions of the tesseract.
 */
tesseract::
tesseract( const vector4 *s )
	: glue(4)
{
	initialize(s);
}

/**
 * Class constructor for a 4-hypercube.
 * @param s the side length of the tesseract.
 */
tesseract::
tesseract( double s )
	: glue(4)
{
	initialize(new vector4(s, s, s, s));
}

/**
 * Class constructor for a standard tesseract.  Each edge will have
 * length 2.
 */
tesseract::
tesseract()
	: glue(4)
{
	initialize(new vector4(2, 2, 2, 2));
}

void tesseract::
initialize( const vector4 *s )
{
	size = *s;
	boundingCucumber(s);

	hyperface[0].coeff[2][0][0][0] = 1;
	hyperface[0].coeff[0][0][0][0] = -s->elem[0]*s->elem[0]/4;

	hyperface[1].coeff[0][2][0][0] = 1;
	hyperface[1].coeff[0][0][0][0] = -s->elem[1]*s->elem[1]/4;

	hyperface[2].coeff[0][0][2][0] = 1;
	hyperface[2].coeff[0][0][0][0] = -s->elem[2]*s->elem[2]/4;

	hyperface[3].coeff[0][0][0][2] = 1;
	hyperface[3].coeff[0][0][0][0] = -s->elem[3]*s->elem[3]/4;

	for (int i = 0; i < 4; ++i)
		add(hyperface+i);

	combine_mode = INTERSECTION;
}

char *tesseract::
toString( char *str )
{
	static char buf[256];

	if (!str) str = buf;
	size.toString(str);

	return str;
}

double tesseract::getHeight(void)  { return size[1]; }
double tesseract::getWidth(void)   { return size[0]; }
double tesseract::getDepth(void)   { return size[2]; }
double tesseract::getTrength(void) { return size[3]; }
