/*
 *  This file is part of Dascant.
 *
 *  Dascant is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Dascant is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Dascant; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
 *  USA
 */

/**
 * A general 4d linear transformation.
 */
#include <transform4.h>
#include <stdio.h>
#include <string.h>

/**
 * Default construct creates the identity transformation.
 */
transform4::
transform4(void)
{
	makeIdentity();
}

/**
 * Copy the data in vector <tt>V</tt> to column <tt>n</tt> of the
 * matrix.
 * @param n the column to update.
 * @param V the vector to take the column from.
 */
void transform4::
setColumn( int n, vector4 &V )
{
	int i;

	for (i = 0; i < 4; ++i)
		matrix[i][n] = V.elem[i];
}

/**
 * Return the <tt>n</tt>th column as a vector.
 * @param n the index of the column to return.
 * @return a vector4 containing the column data.
 */
vector4 transform4::
getColumn( int n )
{
	int i;
	vector4 V;

	for (i = 0; i < 4; ++i)
		V.elem[i] = matrix[i][n];

	return V;
}

/**
 * Copy the data in vector <tt>V</tt> to row <tt>n</tt> of the
 * matrix.
 * @param n the row to update.
 * @param V the vector to take the row from.
 */
void transform4::
setRow( int n, vector4 &V )
{
	matrix[n] = V.elem;
}

/**
 * Return the <tt>n</tt>th row as a vector.
 * @param n the index of the row to return.
 * @return a vector4 &containing the row data.
 */
vector4 transform4::
getRow( int n )
{
	vector4 V(matrix[n]);
	return V;
}

/**
 * Apply this transformation to a vector.
 * @param P the vector4 &to apply the transformation to.
 * @return a vector4 &which is the image of <tt>P</tt> under this
 * transformation.
 */
vector4 transform4::
applyTo( vector4 &P )
{
	int i, j;
	double q[4], *p = P.elem;

	memset(q, 0, sizeof(q));

	for (i = 0; i < 4; ++i)
		for (j = 0; j < 4; ++j)
			q[i] += matrix[i][j] * p[j];

	vector4 V(q);
	return V;
}

/**
 * Apply this transformation to another transformation.
 * @param T the transform4 to apply the transformation to.
 * @return a transform4 which is the composition of this transformation
 * with <tt>T</tt>.
 */
transform4 transform4::
applyTo( transform4 &T )
{
	static transform4 S;
	int i, j, k;

	for (i = 0; i < 4; ++i)
		for (j = 0; j < 4; ++j)
			for (k = 0; k < 4; ++k)
				S.matrix[i][j] += matrix[i][k] * T.matrix[k][j];

	return S;
}

/**
 * Return a string representation of this transformation.  It will take
 * five lines, each of which is a row vector.  It does not end in a
 * newline.
 * @return a String representing this transformation.
 */
char *transform4::
toString( char *s=0x0 )
{
	int i;
	static char buf[256];
	char *cp;
	vector4 V;

	cp = s ? s : buf;

	for (i = 0; i < 4; ++i) {
		cp += sprintf(cp, i ? "  " : "[ ");
		cp += sprintf(cp, "%s %s", (V(matrix[i])).toString(),
			i == 3 ? " ]" : "\n");
	}

	return s ? s : buf;
}

void transform4::
makeIdentity( void )
{
	int i, j;

	for (i = 0; i < 4; ++i)
		for (j = 0; j < 4; ++j)
			matrix[i][j] = i == j;
}
