/*
 *  This file is part of Dascant.
 *
 *  Dascant is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Dascant is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Dascant; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
 *  USA
 */

/**
 * The vector4 class.  This class describes a four-dimensional vector.
 */

#include <ctype.h>
#include <math.h>
#include <stdio.h>
#include <vector4.h>
#include <string.h>

/**
 * Class constructor that duplicates an existing vector.
 * @param V the vector to duplicate.
 */
vector4::
vector4( const vector4 &V )
{
	for (int i = 0; i < 4; ++i)
		elem[i] = V.elem[i];
}

/**
 * Class constructor that allows specification of the coordinates in
 * standard realmspace.
 * @param X the <i>x</i>-coordinate.
 * @param Y the <i>y</i>-coordinate.
 * @param Z the <i>z</i>-coordinate.
 */
vector4::
vector4( double X, double Y, double Z )
{
	elem[0] = X;
	elem[1] = Y;
	elem[2] = Z;
	elem[3] = 0;
}

/**
 * Class constructor that allows specification of the coordinates in
 * tetraspace.
 * @param X the <i>x</i>-coordinate.
 * @param Y the <i>y</i>-coordinate.
 * @param Z the <i>z</i>-coordinate.
 * @param W the <i>w</i>-coordinate.
 */
vector4::
vector4( double X, double Y, double Z, double W )
{
	elem[0] = X;
	elem[1] = Y;
	elem[2] = Z;
	elem[3] = W;
}

/**
 * Class constructor that takes values from an array.  The remaining
 * elements default to the origin's coordinates.
 * @param a an array containing the elements to take.
 * @param n the number of initial elemets in the array to use (default=4).
 */
vector4::
vector4( double a[], int n=4 )
{
	int i;

	for (i = 0; i < 4; ++i)
		if (n > i)
			elem[i] = a[i];
		else
			elem[i] = 0;
}

/**
 * Construct for the origin.
 */
vector4::
vector4( void )
{
	elem[0] = 0;
	elem[1] = 0;
	elem[2] = 0;
	elem[3] = 0;
}

/**
 * Return the coordinates of the vector as a five-element array.
 * @return a five-element array containing the coordinates.
 */
void vector4::
toArray( double a[] )
{
	for (int i = 0; i < 4; ++i)
		a[i] = elem[i];
}

/**
 * Generate a String describing the vector.
 * @return a String.
 */
char *vector4::
toString( char *str=NULL )
{
	static char buf[32];

	if (!str) str = buf;

	sprintf(str, "( %.3g, %.3g, %.3g, %.3g )",
		elem[0], elem[1], elem[2], elem[3]);

	return str;
}

/**
 * Return the L2-norm of the vector.
 * @return the L2-norm of the vector.
 */
double vector4::
norm(void)
{
	double n = sqrt(dot(*this));
	return fabs(n) < 1e-16 ? 0.0 : n;
}

/**
 * Normalize this vector.
 */
void vector4::
normalize(void)
{
	double n = norm();
	for (int i = 0; i < 4; ++i)
		elem[i] /= n;
}

/**
 * Compute the dot product of this vector with another.
 * @param V the other vector in the dot product.
 * @return the dot product.
 */
double vector4::
dot( const vector4 &V ) const
{
	double d = 0.0;
	for (int i = 0; i < 4; ++i)
		d += elem[i] * V.elem[i];
	return fabs(d) < 1e-16 ? 0.0 : d;
}

/**
 * Compute the determinant of the 3x3 matrix described by <tt>U</tt> and
 * <tt>V</tt> with column <tt>exclude</tt> excluded.
 * @param U the vector to use as the second row.
 * @param V the vector to use as the third row.
 * @param exclude the index of the column to exclude.
 * @return the determinant of the matrix generated.
 */
double vector4::
det33( const vector4 &U, const vector4 &V, int exclude )
{
	double d = 0;
	int j[] = { 0, 1, 2 }; /* index to the 3 columns */
	double s[] = { 1, -1, 1, -1 };

	switch(exclude) {
	case 0: j[0]++;
	case 1: j[1]++;
	case 2: j[2]++; }

	d += s[j[0]] * elem[j[0]]
	   * (U.elem[j[1]]*V.elem[j[2]] - U.elem[j[2]]*V.elem[j[1]]);
	d += s[j[1]] * U.elem[j[0]]
	   * (elem[j[1]]*V.elem[j[2]] - elem[j[2]]*V.elem[j[1]]);
	d += s[j[2]] * V.elem[j[0]]
	   * (elem[j[1]]*U.elem[j[2]] - elem[j[2]]*U.elem[j[1]]);

	if (fabs(d) < 1e-16) d = 0.0;

	return d;
}

/**
 * Return the 4 cross product of three vectors.
 * @param U the second vector.
 * @param V the third vector.
 * @return a new vector4 containing the 4 cross product.
 */
vector4 &vector4::
cross( const vector4 &U, const vector4 &V )
{
	static vector4 C;
	double sgn[] = { 1, -1, 1, -1 };

	for (int i = 0; i < 4; ++i)
		C.elem[i] = sgn[i] * det33(U, V, i);

	return C;
}

/**
 * Return a scalar multiple of this vector.
 * @param c the scalar value to multiply.
 * @return the scalar multiple of this vector with <tt>c</tt>.
 */
vector4 vector4::
operator*( double c )
{
	vector4 U(c*elem[0], c*elem[1], c*elem[2], c*elem[3]);
	return U;
}

/**
 * Return a new copy of this vector.
 */
vector4 *vector4::
copy(void)
{
	return new vector4(*this);
}

vector4 vector4::
operator+( vector4 &U )
{
	vector4 v(*this);

	for (int i = 0; i < 4; ++i)
		v[i] += U[i];

	return v;
}

vector4 vector4::
operator=( const vector4 &U )
{
	for (int i = 0; i < 4; ++i)
		elem[i] = U.elem[i];
	return *this;
}

/**
 * Reinitialize this vector from four doubles.
 */
vector4 vector4::
operator()( double x, double y, double z, double w )
{
	elem[0] = x;
	elem[1] = y;
	elem[2] = z;
	elem[3] = w;
	return *this;
}

/**
 * Reinitialize this vector from an array.
 */
vector4 vector4::
operator()( double a[4] )
{
	for (int i = 0; i < 4; ++i)
		elem[i] = a[i];
	return *this;
}

vector4 vector4::
operator+=( vector4 &U )
{
	*this = *this + U;
	return *this;
}

int vector4::
fromString( const char *str )
{
	while (*str && isspace(*str)) str++;
	int r = sscanf(str, "( %lf, %lf, %lf, %lf )",
		elem+0, elem+1, elem+2, elem+3);
	return r;
}
