/**
 * The Camera class.  A Camera keeps track of its own rays, for the purpose
 * of raytracing.  Repositioning the camera causes new rays to be generated
 * (but not traced).
 */
public class Camera {
	public Vector4D position;
	public Vector4D axis[];
	public Ray ray[][];
	public int width, height;
	public Transform4D orientation;
	public double depth;

	/**
	 * Class constructor for camera with default values.
	 */
	public Camera()
	{
		initialize(128, 128, 200);
	}

	/**
	 * Class constructor.
	 * @param w width of the image plane
	 * @param h height of the image plane
	 * @param d image plane depth (distance from the camera center).
	 */
	public Camera( int w, int h, double d )
	{
		initialize(w, h, d);
	}

	private void initialize( int w, int h, double d )
	{
		position = new Vector4D();

		axis = new Vector4D[2];
		axis[0] = new Vector4D(0, 0, -1, 0);
		axis[1] = new Vector4D(0, 0, 0, 1);
		try { orientation = new Rotation4D(axis[0], axis[1]); }
		catch(GeometryException e) {
			e.printStackTrace();
			System.exit(1);
		}

		width = w;
		height = h;
		depth = d;

		ray = new Ray[height][width];
	}

	/**
	 * Create a copy of this camera.
	 */
	public Camera copy()
	{
		Camera newCam = new Camera(width, height, depth);
		newCam.position = position.copy();
		newCam.axis = new Vector4D[2];
		newCam.axis[0] = axis[0].copy();
		newCam.axis[1] = axis[1].copy();
		newCam.orientation = orientation.copy();
		newCam.height = height;
		newCam.width = width;
		newCam.ray = new Ray[height][width];
		for (int i = 0; i < height; ++i)
			for (int j = 0; j < width; ++j)
				newCam.ray[i][j] = ray[i][j].copy(this,
					j-width/2, i-height/2);
		return newCam;
	}

	public Vector4D getPosition()
	{
		return position.copy();
	}

	public double getPosition( int m )
	{
		return position.elem[m];
	}

	public void setPosition( double x, double y, double z, double w )
	{
		position = new Vector4D(x, y, z, w);
		compute_rays();
	}

	public void setPosition( int m, double f )
	{
		position.elem[m] = f;
		compute_rays();
	}

	public void setPosition( Vector4D V )
	{
		position = V.copy();
		compute_rays();
	}

	public void compute_rays()
	{
		int i, j;

		for (i = 0; i < height; ++i)
			for (j = 0; j < width; ++j)
				ray[i][j] = new Ray(this, j-width/2, i-height/2);
	}
};
