/**
 * The DSML_Lexer class.  A lexical analyzer for DSML files.
 */

import java.io.*;
import java.util.Stack;
import java.util.StringTokenizer;

public class DSML_Lexer {
	/**
	 * A string representing the version number for this file.
	 */
	public String version;

	/**
	 * A flag indicating whether the end of the file has been reached.
	 */
	public boolean eof;

	/**
	 * The most recently eaten token.
	 */
	public DSML_Token currentToken;

	private StringTokenizer tokenizer;
	private Stack<DSML_Token> tokenStack;
	private BufferedReader reader;
	private static final String delims = " \t\n;,=(){}[]";
	private int line_number;

	/**
	 * Class constructor to start analyzing a new file.
	 * @param filename the name of the file to analyze.
	 */
	public DSML_Lexer( String filename )
		throws Exception
	{
		FileInputStream is = new FileInputStream(filename);
		reader = new BufferedReader(new InputStreamReader(is));

		DSML_Exception.registerLexer(this);

		/* check version number */
		tokenizer = new StringTokenizer(reader.readLine());
		if (tokenizer.countTokens() < 2)
			throw new DSML_Exception("no DSML header found");

		String s = tokenizer.nextToken();
		if (s.compareTo("DSML") != 0)
			throw new DSML_Exception("expected DSML header; got \""
				+ s + "\"");
		version = tokenizer.nextToken();

		eof = false;

		line_number = 1;

		tokenStack = new Stack<DSML_Token>();
	}

	/**
	 * Get the next token in the file.  Also sets <tt>currentToken</tt>.
	 * @return a new DSML_Token for the current token.
	 */
	public DSML_Token nextToken()
	{
		String s;
		char c;

		if (!tokenStack.empty()) {
			currentToken = tokenStack.pop();
			return currentToken;
		}

		currentToken = null;

		try {
			while (tokenizer==null || !tokenizer.hasMoreTokens()) {
				tokenizer = new StringTokenizer(reader.readLine(),
					delims, true);
				line_number++;
			}
		} catch (Exception e) {
			eof = true;
			return new DSML_Token(DSML_Token.TokenType.TT_EOF);
		}

		try {
			do { s = tokenizer.nextToken(); }
			while (s.charAt(0)==' ' || s.charAt(0)=='\t'
				|| s.charAt(0)=='\n');
		} catch(Exception e) {
			System.err.println(e.getMessage());
			return null;
		}

		if (s.compareTo("#") == 0) {
			tokenizer = null;
			return nextToken();
		}

		c = s.charAt(0);

		/* tokenizer.ttype = TT_WORD; determine what kind */
		switch(c) {
		case ';':
			currentToken = new DSML_Token(DSML_Token.TokenType.TT_Semicolon);
			break;
		case ',':
			currentToken = new DSML_Token(DSML_Token.TokenType.TT_Comma);
			break;
		case '=':
			currentToken = new DSML_Token(DSML_Token.TokenType.TT_Equals);
			break;
		case '+':
			currentToken = new DSML_Token(DSML_Token.TokenType.TT_Plus);
			break;
		case '-':
			currentToken = new DSML_Token(DSML_Token.TokenType.TT_Minus);
			break;
		case '{':
			currentToken = new DSML_Token(DSML_Token.TokenType.TT_LeftCurly);
			break;
		case '}':
			currentToken = new DSML_Token(DSML_Token.TokenType.TT_RightCurly);
			break;
		case '[':
			currentToken = new DSML_Token(DSML_Token.TokenType.TT_LeftSquare);
			break;
		case ']':
			currentToken = new DSML_Token(DSML_Token.TokenType.TT_RightSquare);
			break;
		case '(':
			currentToken = new DSML_Token(DSML_Token.TokenType.TT_LeftParenthesis);
			break;
		case ')':
			currentToken = new DSML_Token(DSML_Token.TokenType.TT_RightParenthesis);
			break;
		}

		if (currentToken != null)
			return currentToken;

		/* must be an identifier or a keyword */
		if (s.compareTo("shape") == 0)
			currentToken = new DSML_Token(DSML_Token.TokenType.TT_KW_Shape);
		else if (s.compareTo("instance") == 0)
			currentToken = new DSML_Token(DSML_Token.TokenType.TT_KW_Instance);
		else if (s.compareTo("of") == 0)
			currentToken = new DSML_Token(DSML_Token.TokenType.TT_KW_Of);
		else if (s.compareTo("box") == 0)
			currentToken = new DSML_Token(DSML_Token.TokenType.TT_KW_Box);
		else if (s.compareTo("axis") == 0)
			currentToken = new DSML_Token(DSML_Token.TokenType.TT_KW_Axis);
		else if (s.compareTo("color") == 0)
			currentToken = new DSML_Token(DSML_Token.TokenType.TT_KW_Color);
		else if (s.compareTo("glue") == 0)
			currentToken = new DSML_Token(DSML_Token.TokenType.TT_KW_Glue);
		else if (s.compareTo("transform") == 0)
			currentToken = new DSML_Token(DSML_Token.TokenType.TT_KW_Transform);
		else if (s.compareTo("scale") == 0)
			currentToken = new DSML_Token(DSML_Token.TokenType.TT_KW_Scale);
		else if (s.compareTo("along") == 0)
			currentToken = new DSML_Token(DSML_Token.TokenType.TT_KW_Along);
		else if (s.compareTo("rotate") == 0)
			currentToken = new DSML_Token(DSML_Token.TokenType.TT_KW_Rotate);
		else if (s.compareTo("around") == 0)
			currentToken = new DSML_Token(DSML_Token.TokenType.TT_KW_Around);
		else if (s.compareTo("translate") == 0)
			currentToken = new DSML_Token(DSML_Token.TokenType.TT_KW_Translate);
		else if (s.compareTo("by") == 0)
			currentToken = new DSML_Token(DSML_Token.TokenType.TT_KW_By);
		else if (s.compareTo("clip") == 0)
			currentToken = new DSML_Token(DSML_Token.TokenType.TT_KW_Clip);
		else if (s.compareTo("to") == 0)
			currentToken = new DSML_Token(DSML_Token.TokenType.TT_KW_To);
		else if (s.compareTo("cut") == 0)
			currentToken = new DSML_Token(DSML_Token.TokenType.TT_KW_Cut);
		else if (s.compareTo("out") == 0)
			currentToken = new DSML_Token(DSML_Token.TokenType.TT_KW_Out);
		else if (s.compareTo("viewpoint") == 0)
			currentToken = new DSML_Token(DSML_Token.TokenType.TT_KW_Out);

		/* must be an identifier or a number or a variable */
		else if (c == '+' || c == '-' || '0' <= c && c <= '9')
			currentToken = new DSML_Token(Double.valueOf(s).doubleValue());
		else if (c == 'x' || c == 'y' || c == 'z' || c == 'w')
			currentToken = DSML_Token.newVariableToken(s);

		if (currentToken == null)
			currentToken = new DSML_Token(s);

		return currentToken;
	}

	/**
	 * Get the current line number.
	 * @return the current line number.
	 */
	public int lineno()
	{
		return line_number;
	}

	/**
	 * Unget the current token.
	 */
	public void pushBack()
	{
		tokenStack.push(currentToken);
		currentToken = null;
	}
};
