/**
 * The DSML_Parser class.  A parser for DSML files.
 */
public class DSML_Parser {
	private DSML_Lexer lex;

	private void parse_scene()
		throws DSML_Exception
	{
		while (!lex.eof)
			parse_definition();
	}

	private void parse_definition()
		throws DSML_Exception
	{
		DSML_Token t;

		t = lex.nextToken();
		switch(t.type) {
		case TT_KW_Shape: parse_shape(); break;
		case TT_KW_Viewpoint: parse_viewpoint(); break;
		case TT_KW_Color: parse_color(); break;
		case TT_KW_Glue: parse_glue(); break;
		case TT_KW_Transform: parse_transform(); break;
		case TT_KW_Instance: parse_instance(); break;
		case TT_EOF: break;
		default:
			throw new DSML_Exception(t + " cannot begin a definition.");
		}

	}

	private void parse_instance()
		throws DSML_Exception
	{
		DSML_Token t, name, shape;

		name = lex.nextToken();
		if (name.type != DSML_Token.TokenType.TT_Identifier)
			throw new DSML_Exception("identifier for name of instance", name);

		t = lex.nextToken();
		if (t.type != DSML_Token.TokenType.TT_KW_Of)
			throw new DSML_Exception("keyword \"of\"", t);

		shape = lex.nextToken();
		if (shape.type != DSML_Token.TokenType.TT_Identifier)
			throw new DSML_Exception("identifier for name of shape", shape);

		t = lex.nextToken();
		if (t.type != DSML_Token.TokenType.TT_LeftCurly)
			throw new DSML_Exception("{", t);

		parse_shape_body();
	}

	private void parse_shape()
		throws DSML_Exception
	{
		DSML_Token t, name;

		name = lex.nextToken();
		if (name.type != DSML_Token.TokenType.TT_Identifier)
			throw new DSML_Exception("identifier for name of shape", name);

		t = lex.nextToken();
		if (t.type != DSML_Token.TokenType.TT_LeftCurly)
			throw new DSML_Exception("{ to begin shape body", t);

		parse_shape_body();
	}

	private void parse_shape_body()
		throws DSML_Exception
	{
		while (parse_shape_element());
	}

	private boolean parse_shape_element()
		throws DSML_Exception
	{
		DSML_Token t;

		t = lex.nextToken();

		switch(t.type) {
		case TT_RightCurly: return false;
		case TT_KW_Color: parse_color_spec(); break;
		case TT_KW_Axis: parse_axis_spec(); break;
		case TT_KW_Box: parse_box_spec(); break;
		default:
			lex.pushBack();
			Quadric Q = parse_quadric();
			System.out.println(Q);
			break;
		}

		return true;
	}

	private void parse_color_spec()
		throws DSML_Exception
	{
	}

	private void parse_axis_spec()
		throws DSML_Exception
	{
	}

	private Vector4D parse_vector()
		throws DSML_Exception
	{
		DSML_Token t, elem;
		int i;
		double v[] = new double[4];

		t = lex.nextToken();
		if (t.type != DSML_Token.TokenType.TT_LeftSquare)
			throw new DSML_Exception("expeceted [");

		for (i = 0; i < 4; ++i) {
			elem = lex.nextToken();
			if (elem.type != DSML_Token.TokenType.TT_Number)
				throw new DSML_Exception("expeceted number");

			v[i] = elem.number;

			t = lex.nextToken();
			if (i == 3) {
				if (t.type != DSML_Token.TokenType.TT_RightSquare)
					throw new DSML_Exception("expeceted ]");
			} else if (t.type != DSML_Token.TokenType.TT_Comma)
				throw new DSML_Exception("expeceted ,");
		}

		return new Vector4D(v);
	}

	private void parse_box_spec()
		throws DSML_Exception
	{
		DSML_Token t;

		parse_vector();

		t = lex.nextToken();
		if (t.type != DSML_Token.TokenType.TT_Semicolon)
			throw new DSML_Exception(";", t);
	}

	private Quadric parse_quadric()
		throws DSML_Exception
	{
		DSML_Token t, coeff, var, ex, op, q2;
		Quadric Q = new Quadric();
		Term T;
		double sign = 1.0;
		boolean gotone = false;
		boolean go = true;

		while(go) {
			T = parse_term();
			gotone = true;

			T.coeff *= sign;
			try { Q.add(T); }
			catch(GeometryException e) {
				throw new DSML_Exception(e.getMessage());
			}

			t = lex.nextToken();
			switch(t.type) {
			case TT_Semicolon: go = false; break;
			case TT_Plus:  sign = +1.0; break;
			case TT_Minus: sign = -1.0; break;
			default: throw new DSML_Exception("+ or - or ;", t);
			}
		}

		return Q;
	}

	private Term parse_term()
		throws DSML_Exception
	{
		DSML_Token t, var, ex;
		Term T = new Term();
		boolean go = true;

		t = lex.nextToken();

		if (t.type == DSML_Token.TokenType.TT_Number) {
			T.coeff = t.number;
			var = lex.nextToken();
		} else {
			T.coeff = 1.0;
			var = t;
		}

		while(go) {
			if (var.type != DSML_Token.TokenType.TT_Variable) {
				lex.pushBack();
				break;
			}

			switch(var.variable) {
			case VarX:  T.powX += 1; break;
			case VarY:  T.powY += 1; break;
			case VarZ:  T.powZ += 1; break;
			case VarW:  T.powW += 1; break;
			case VarX2: T.powX += 2; break;
			case VarY2: T.powY += 2; break;
			case VarZ2: T.powZ += 2; break;
			case VarW2: T.powW += 2; break;
			}

			var = lex.nextToken();
		}

		return T;
	}

	private void parse_viewpoint()
		throws DSML_Exception
	{
	}

	private void parse_color()
		throws DSML_Exception
	{
	}

	private void parse_glue()
		throws DSML_Exception
	{
		DSML_Token name, t;

		name = lex.nextToken();
		if (name.type != DSML_Token.TokenType.TT_Identifier)
			throw new DSML_Exception("identifier", name);

		t = lex.nextToken();
		if (t.type != DSML_Token.TokenType.TT_LeftCurly)
			throw new DSML_Exception("{", t);

		while(parse_glue_component());
	}

	private boolean parse_glue_component()
		throws DSML_Exception
	{
		DSML_Token t;

		t = lex.nextToken();
		if (t.type == DSML_Token.TokenType.TT_RightCurly)
			return false;

		if (t.type == DSML_Token.TokenType.TT_LeftParenthesis)
			parse_application();

		t = lex.nextToken();
		if (t.type != DSML_Token.TokenType.TT_Semicolon)
			throw new DSML_Exception(";", t);

		return true;
	}

	private void parse_application()
		throws DSML_Exception
	{
		DSML_Token t, s, head, variable, input=null;

		head = lex.nextToken();
		if (head.type != DSML_Token.TokenType.TT_Identifier)
			throw new DSML_Exception("identifier as head of application",
				head);

		t = lex.nextToken();
		/* expect a list of bindings, followed by a ) */
		/* invariant: t is an identifier not preceeded by an =,
		 *         or t is a ),
		 *         or t is a ( and an application follows */
		while (true) {
			if (t.type == DSML_Token.TokenType.TT_LeftParenthesis) {
				parse_application();
				break;
			}

			if (t.type == DSML_Token.TokenType.TT_RightParenthesis)
				throw new DSML_Exception("haven't gotten input yet.");

			if (t.type != DSML_Token.TokenType.TT_Identifier)
				throw new DSML_Exception("identifier for binding or input",
					t);

			/* is this the input? */
			s = lex.nextToken();
			if (s.type == DSML_Token.TokenType.TT_RightParenthesis)
				return;

			/* is this a variable binding? */
			if (s.type == DSML_Token.TokenType.TT_Equals) {
				variable = t;
				parse_quantity();
				s = lex.nextToken();
				if (s.type != DSML_Token.TokenType.TT_Comma)
					throw new DSML_Exception(", after binding", s);
				t = lex.nextToken();
				if (t.type != DSML_Token.TokenType.TT_Identifier)
					throw new DSML_Exception("application input", t);
				continue;
			}
		} /* while (true) */

		t = lex.nextToken();
		if (t.type != DSML_Token.TokenType.TT_RightParenthesis)
			throw new DSML_Exception(")", t);
	}

	private void parse_transform()
		throws DSML_Exception
	{
		DSML_Token t, name;

		name = lex.nextToken();
		if (name.type != DSML_Token.TokenType.TT_Identifier)
			throw new DSML_Exception("identifier", name);

		t = lex.nextToken();
		if (t.type != DSML_Token.TokenType.TT_LeftCurly)
			throw new DSML_Exception("{", t);

		while(parse_transform_element());
	}

	private boolean parse_transform_element()
		throws DSML_Exception
	{
		DSML_Token t;

		t = lex.nextToken();
		switch(t.type) {
		case TT_RightCurly: return false;
		case TT_KW_Scale: parse_scale_defn(); break;
		case TT_KW_Rotate: parse_rotate_defn(); break;
		case TT_KW_Translate: parse_translate_defn(); break;
		case TT_KW_Clip: parse_clip_defn(); break;
		case TT_KW_Cut: parse_cut_defn(); break;
		default: throw new DSML_Exception("unknown transformation " + t);
		}

		return true;
	}

	private void parse_scale_defn()
		throws DSML_Exception
	{
		DSML_Token t, quantity, axis;

		parse_quantity();

		t = lex.nextToken();
		if (t.type == DSML_Token.TokenType.TT_Semicolon)
			return;

		if (t.type != DSML_Token.TokenType.TT_KW_Along)
			throw new DSML_Exception("along or ;", t);

		parse_axis();

		t = lex.nextToken();
		if (t.type != DSML_Token.TokenType.TT_Semicolon)
			throw new DSML_Exception(";", t);
	}

	private void parse_axis()
		throws DSML_Exception
	{
		DSML_Token t;

		throw new DSML_Exception("axes not yet implemented");
	}

	private void parse_quantity()
		throws DSML_Exception
	{
		DSML_Token t;

		t = lex.nextToken();
		if (t.type != DSML_Token.TokenType.TT_Number
		 && t.type != DSML_Token.TokenType.TT_Identifier)
			throw new DSML_Exception("number or identifier", t);
	}

	private void parse_rotate_defn()
		throws DSML_Exception
	{
	}

	private void parse_translate_defn()
		throws DSML_Exception
	{
		DSML_Token t;

		t = lex.nextToken();
		if (t.type != DSML_Token.TokenType.TT_KW_By)
			throw new DSML_Exception("by", t);

		parse_vector();

		t = lex.nextToken();
		if (t.type != DSML_Token.TokenType.TT_Semicolon)
			throw new DSML_Exception(";", t);
	}

	private void parse_clip_defn()
		throws DSML_Exception
	{
	}

	private void parse_cut_defn()
		throws DSML_Exception
	{
		throw new DSML_Exception("cut is not implenented.");
	}

	/**
	 * Class constructor.
	 * @param filename the name of the file to parse.
	 */
	public DSML_Parser( String filename )
		throws Exception
	{
		lex = new DSML_Lexer(filename);
	}

	/**
	 * Start parsing the file.
	 */
	public void parse()
		throws DSML_Exception
	{
		parse_scene();
		System.out.println("Parse successful.");
	}
};
