/**
 * The DSML_Token class.  This class represents a token in a DSML file.
 */
public class DSML_Token {
	/**
	 * The TokenType enum lists all possible kinds of tokens.
	 */
	public enum TokenType {
		TT_LeftCurly, TT_RightCurly,
		TT_LeftParenthesis, TT_RightParenthesis,
		TT_LeftSquare, TT_RightSquare,
		TT_Semicolon, TT_Comma, TT_Equals, TT_Plus, TT_Minus,
		TT_Identifier, TT_Number, TT_Variable,
		TT_KW_Shape, TT_KW_Instance, TT_KW_Axis, TT_KW_Color,
		TT_KW_Of, TT_KW_Box, TT_KW_Glue,
		TT_KW_Transform,
		TT_KW_Scale, TT_KW_Along,
		TT_KW_Rotate, TT_KW_Around,
		TT_KW_Translate, TT_KW_By,
		TT_KW_Clip, TT_KW_To,
		TT_KW_Cut, TT_KW_Out,
		TT_KW_Viewpoint,
		TT_EOF
	};

	/**
	 * The Variable enum is used to represent a set of coordinate variables
	 * when generating a formula for shapes.
	 */
	public enum Variable {
		VarX, VarY, VarZ, VarW,
		VarX2, VarY2, VarZ2, VarW2
	};

	/**
	 * This token's type.
	 */
	public TokenType type;

	/**
	 * If this token is a variable (TT_Variable), this will hold the value.
	 */
	public Variable variable;

	/**
	 * If this token is a number (TT_Number), this will hold the value.
	 */
	public double number;

	/**
	 * If this token is an identifier (TT_Identifier), this will hold the
	 * symbol.
	 */
	public String identifier;

	/**
	 * Set up the initial values for the fields in this token.
	 * @param t the type of token.
	 */
	private void token_init( TokenType t )
	{
		type = t;
		variable = Variable.VarX;
		number = 0.0;
		identifier = null;
	}

	/**
	 * Class constructor for a token with no special value.
	 * @param t the token type.
	 */
	public DSML_Token( TokenType t )
	{
		token_init(t);
	}

	/**
	 * Class constructor for a variable-type token.
	 * @param v the variable this token represents.
	 */
	public DSML_Token( Variable v )
	{
		token_init(TokenType.TT_Variable);
		variable = v;
	}

	/**
	 * Class constructor for a numeric token.
	 * @param x the value of this token.
	 */
	public DSML_Token( double x )
	{
		token_init(TokenType.TT_Number);
		number = x;
	}

	/**
	 * Class constructor for an identifier token.
	 * @param id the symbol.
	 */
	public DSML_Token( String id )
	{
		token_init(TokenType.TT_Identifier);
		identifier = id;
	}

	/**
	 * Create a new DSML_Token object of type TT_Variable, where the
	 * variable is derived from a string.  A variable is a product of X, Y,
	 * Z, and W, in alphabetical order.
	 * @param s the string to get the variable name from.
	 * @return a new DSML_Token of type TT_Variable.
	 */
	public static DSML_Token newVariableToken( String s )
	{
		     if (s.compareTo("x")==0) 
			return new DSML_Token(Variable.VarX);
		else if (s.compareTo("y")==0) 
			return new DSML_Token(Variable.VarY);
		else if (s.compareTo("z")==0) 
			return new DSML_Token(Variable.VarZ);
		else if (s.compareTo("w")==0) 
			return new DSML_Token(Variable.VarW);

		     if (s.compareTo("x2")==0) 
			return new DSML_Token(Variable.VarX2);
		else if (s.compareTo("y2")==0) 
			return new DSML_Token(Variable.VarY2);
		else if (s.compareTo("z2")==0) 
			return new DSML_Token(Variable.VarZ2);
		else if (s.compareTo("w2")==0) 
			return new DSML_Token(Variable.VarW2);

		return null;
	};

	/**
	 * Return a string representation of this variable.
	 * @return a string representing this variable.
	 */
	public String toString()
	{
		String t="", d=null;

		switch(type) {
		case TT_LeftCurly: t = "LeftCurly"; break;
		case TT_RightCurly: t = "RightCurly"; break;
		case TT_LeftParenthesis: t = "LeftParenthesis"; break;
		case TT_RightParenthesis: t = "RightParenthesis"; break;
		case TT_LeftSquare: t = "LeftSquare"; break;
		case TT_RightSquare: t = "RightSquare"; break;
		case TT_Semicolon: t = "Semicolon"; break;
		case TT_Comma: t = "Comma"; break;
		case TT_Equals: t = "Equals"; break;
		case TT_Plus: t = "Plus"; break;
		case TT_Minus: t = "Minus"; break;
		case TT_Identifier:
			t = "Identifier";
			d = identifier;
			break;
		case TT_Number:
			t = "Number";
			d = "" + number;
			break;
		case TT_Variable:
			t = "Variable";
			switch(variable) {
			case VarX: d = "x"; break;
			case VarY: d = "y"; break;
			case VarZ: d = "z"; break;
			case VarW: d = "w"; break;
			case VarX2: d = "x^2"; break;
			case VarY2: d = "y^2"; break;
			case VarZ2: d = "z^2"; break;
			case VarW2: d = "W^2"; break;
			}
			break;
		case TT_KW_Shape: t = "KW_Shape"; break;
		case TT_KW_Instance: t = "KW_Instance"; break;
		case TT_KW_Of: t = "KW_Of"; break;
		case TT_KW_Box: t = "KW_Box"; break;
		case TT_KW_Axis: t = "KW_Axis"; break;
		case TT_KW_Color: t = "KW_Color"; break;
		case TT_KW_Glue: t = "KW_Glue"; break;
		case TT_KW_Transform: t = "KW_Transform"; break;
		case TT_KW_Scale: t = "KW_Scale"; break;
		case TT_KW_Along: t = "KW_Along"; break;
		case TT_KW_Rotate: t = "KW_Rotate"; break;
		case TT_KW_Around: t = "KW_Around"; break;
		case TT_KW_Translate: t = "KW_Translate"; break;
		case TT_KW_By: t = "KW_By"; break;
		case TT_KW_Clip: t = "KW_Clip"; break;
		case TT_KW_To: t = "KW_To"; break;
		case TT_KW_Cut: t = "KW_Cut"; break;
		case TT_KW_Out: t = "KW_Out"; break;
		case TT_KW_Viewpoint: t = "KW_Viewpoint"; break;
		case TT_EOF: t = "EOF"; break;
		}

		return "<" + t + (d==null ? "" : (" " + d)) + ">";
	}
};
