/**
 * The Quadric class.  Objects of this type represent four-dimenional
 * quadric surfaces.
 */

import java.text.NumberFormat;

public class Quadric extends Entity {
	/**
	 * List of terms.  Each element is an array of exponents of x, y, z, and
	 * w.
	 */
	public static final int term[][] = {
		{ 2, 0, 0, 0 }, // square terms
		{ 0, 2, 0, 0 },
		{ 0, 0, 2, 0 },
		{ 0, 0, 0, 2 },
		{ 1, 1, 0, 0 }, // cross terms
		{ 1, 0, 1, 0 },
		{ 1, 0, 0, 1 },
		{ 0, 1, 1, 0 },
		{ 0, 1, 0, 1 },
		{ 0, 0, 1, 1 },
		{ 1, 0, 0, 0 }, // linear terms
		{ 0, 1, 0, 0 },
		{ 0, 0, 1, 0 },
		{ 0, 0, 0, 1 },
		{ 0, 0, 0, 0 }, // constant term
	};

	/**
	 * Degree of terms.
	 */
	public static final int degree[] = {
		2, 2, 2, 2,
		2, 2, 2, 2, 2, 2,
		1, 1, 1, 1,
		0
	};

	/**
	 * The coefficient matrix.  <tt>coeff[i][j][k][l]</tt> is the
	 * coefficient of <i>x<sup>i</sup> y<sup>j</sup> z<sup>k</sup>
	 * w<sup>l</sup></i>.
	 */ 
	public double coeff[][][][];

	/**
	 * Surface orientation.  The form of the quadric solid is &nu;P(x,y,z,w)
	 * &lt;= 0.  If this value is 1, then, the surface will be oriented in
	 * the normal way.  Otherwise, the solid described will be the
	 * complement of the standard solid you'd expect (such as an ellipsonkal
	 * hole instead of an ellipsonk).
	 */
	public int nu;

	/**
	 * Class constructor; generate the "zero" quadric.
	 */
	public Quadric()
	{
		super();

		int x, y, z, w;

		coeff = new double[3][3][3][3];

		for (x = 0; x <= 2; ++x)
			for (y = 0; y <= 2; ++y)
				for (z = 0; z <= 2; ++z)
					for (w = 0; w <= 2; ++w)
						coeff[x][y][z][w] = 0.0;

		nu = 1;
	}

	/**
	 * Return the coefficient of the <tt>trm</tt>th term.
	 * @param trm a Term object used to index the coefficient array.
	 * @return the coefficient of the index term.
	 */
	public double getCoefficient( Term trm )
		throws GeometryException
	{
		if (trm.degree() > 2)
			throw new GeometryException("For quadric surfaces, "
			 + "no term can have degree greater than 2, but this "
			 + "term is " + trm);
		return coeff[trm.powX][trm.powY][trm.powZ][trm.powW];
	}

	/**
	 * Set the coefficient of the <tt>trm</tt>th term to <tt>c</tt>.
	 * @param c the new coefficient.
	 * @param trm an index to the desired term.
	 */
	public void setCoefficient( double c, Term trm )
		throws GeometryException
	{
		if (trm.degree() > 2)
			throw new GeometryException("For quadric surfaces, "
			 + "no term can have degree greater than 2, but this "
			 + "term is " + trm);
		coeff[trm.powX][trm.powY][trm.powZ][trm.powW] = c;
	}

	/**
	 * Add term <tt>T</tt> to this quadric.
	 * @param T the term to add.
	 */
	public void add( Term T ) 
		throws GeometryException
	{
		if (T.degree() > 2)
			throw new GeometryException("For quadric surfaces, "
			 + "no term can have degree greater than 2, but this "
			 + "term is " + T);
		coeff[T.powX][T.powY][T.powZ][T.powW] += T.coeff;
	}

	/**
	 * Detect collisions.  Do this by substituting the parametrized form of
	 * the ray into the equation for the surface to derive a quadric
	 * equation in <i>t</i>, the parameter for the ray.  Solve this
	 * equation, and use the solutions to determine all collisions with the
	 * surface.
	 * @param R a Ray to check for collisions with.
	 * @return a Collision object, or <tt>null</tt> if no collisions occur.
	 */
	public Collision[] detectCollisions( Ray R )
	{
		double a, b, c, desc, soln;
		int i, j;
		Collision C[];

		// find quadratic equation  a t^2 + b t + c = 0  by substituting
		// parametric form of Ray into this quadric.  Ray conveniently
		// provides us with the coefficients of t^2, t, and 1 for each
		// possible product of x, y, z, and w, so we only need to multiply
		// those products by the coefficients of the terms in the equation
		// of the surface.
		a = b = c = 0;
		for (i = 0; i < term.length; ++i) {
			a += nu * coeff[term[i][0]][term[i][1]][term[i][2]][term[i][3]]
			        * R.ttab[i][2];
			b += nu * coeff[term[i][0]][term[i][1]][term[i][2]][term[i][3]]
			        * R.ttab[i][1];
			c += nu * coeff[term[i][0]][term[i][1]][term[i][2]][term[i][3]]
			        * R.ttab[i][0];
		}

		// there has to be at least one term in the quadtratic equation with
		// a t in it.
		if (Math.abs(a) < 1e-16 && Math.abs(b) < 1e-16)
			return null;

		desc = b*b - 4*a*c;
		if (desc < 0) return null;

		double sqrtDesc = Math.sqrt(desc);
		if (desc < 1e-16) {
			C = new Collision[1];
			C[0] = new Collision(R, -b/(2*a), this);
		} else {
			C = new Collision[2];
			C[0] = new Collision(R, (-b - sqrtDesc) / (2*a), this);
			C[1] = new Collision(R, (-b + sqrtDesc) / (2*a), this);
		}
		/* TODO: Figure out outgoing ray, insideness. */

		/* surface normals */
		/* FIXME: only takes into account the square terms. */
		double sq[] = new double[4];
		sq[0] = coeff[2][0][0][0];
		sq[1] = coeff[0][2][0][0];
		sq[2] = coeff[0][0][2][0];
		sq[3] = coeff[0][0][0][2];
		for (i = 0; i < C.length; ++i) {
			C[i].normal = new Vector4D();
			for (j = 0; j < 4; ++j) {
				// consider square term
				C[i].normal.elem[j] = 2 * nu * C[i].location.elem[j] * sq[j];
				//  consider cross terms
				for (int k = 0; k < 4; ++k)
					if (j != k)
						C[i].normal.elem[j] += nu * C[i].location.elem[k]
							* coeff[j==0||k==0?1:0]
							       [j==1||k==1?1:0]
							       [j==2||k==2?1:0]
							       [j==3||k==3?1:0];
				// Consider linear term
				C[i].normal.elem[j] += nu *
					coeff[j==0?1:0][j==1?1:0][j==2?1:0][j==3?1:0];
			}
			C[i].normal.normalize();
		}

		return C;
	}

	/**
	 * Add to this quadric another.  Simply adds corresponding coefficients.
	 * @param Q the other quadric to add to this one.
	 */
	public void add( Quadric Q )
	{
		int i, j, k, l;

		for (i = 0; i <= 2; ++i)
			for (j = 0; j <= 2-i; ++j)
				for (k = 0; k <= 2-i-j; ++k)
					for (l = 0; l <= 2-i-j-k; ++l)
						coeff[i][j][k][l] += Q.coeff[i][j][k][l];
	}

	/**
	 * Return, as a string, the equation defining this quadric surface.
	 * @return the equation as a string.
	 */
	public String toString()
	{
		int i, j, k, l;
		String s;
		double c;
		boolean first=true;
		NumberFormat fmt = NumberFormat.getInstance();

		fmt.setMaximumFractionDigits(4);

		s = "";
		for (i = 2; i >= 0; --i)
			for (j = 2-i; j >= 0; --j)
				for (k = 2-i-j; k >= 0; --k)
					for (l = 2-i-j-k; l >= 0; --l) {
						if (i+j+k+l == 0) continue;

						c = coeff[i][j][k][l];
						if (Math.abs(c) < 1e-16)
							continue;

						if (c < 0)                s += "- ";
						else if (c > 0 && !first) s += "+ ";
						first = false;

						if (Math.abs(c-1) > 1e-16)
							s += fmt.format(Math.abs(c));

						if (i > 0) { s += "x"; if (i > 1) s += "^2 "; }
						if (j > 0) { s += "y"; if (j > 1) s += "^2 "; }
						if (k > 0) { s += "z"; if (k > 1) s += "^2 "; }
						if (l > 0) { s += "w"; if (l > 1) s += "^2 "; }
					}

		return s
		     + (nu < 0 ? ">= " : "<= ")
		     + fmt.format(-coeff[0][0][0][0]);
	}
};
