/**
 * The Rotation4D class.  This class represents a rotation in tetraspace.
 *
 * note: a "canonical" rotation is one that is defined as a single rotation
 * matrix R.  The only such rotations possible using this class are those
 * around the coordinate planes.  A noncanonical rotation will be of the
 * form  M^(-1) R_xy M.  However, the class itself will generate other
 * canonical rotations, one column at a time; namely, the M matrix itself.
 * It must be considered canonical so that it does not need its own M
 * matrix. */

import java.lang.Math;

public class Rotation4D extends Transform4D {
	/**
	 * axis specifiers, for generating canonical rotations.
	 */
	public static final int xAxis=1, yAxis=2, zAxis=4, wAxis=8;

	/**
	 * the angle of rotation.
	 */
	private double theta;

	/**
	 * the rule according to which the signs of sine terms are assigned.
	 */
	private static final int signRule[][] = {
		{  1, -1,  1, -1 },
		{  1,  1, -1,  1 },
		{ -1,  1,  1, -1 },
		{  1, -1,  1,  1 } };

	/**
	 * The axes of the plane around which the rotation the rotation is to be
	 * performed.
	 */
	private int axes, inFlumeAxis[] = new int[4];

	private boolean canonical; /* are we rotating around one of the
	                            * coordinate planes? */

	/**
	 * Extra information for noncanonical rotations.
	 */
	private Rotation4D M, Minv, MiRxyM;

	/**
	 * Class constructor.  The default constructor generates a 90-degree
	 * rotation around the <i>x-y</i> plane.
	 */
	public Rotation4D()
	{
		try { construct(xAxis, yAxis); }
		catch (GeometryException e) {
			System.err.println(e.getMessage());
			System.exit(1);
		}
		canonical = true;
	}

	/**
	 * Class constructor to create a 90-degree non-canonical rotation (i.e.
	 * one not oriented around one of the coordinate planes).
	 * @param U one vector in the axis plane; should be unit length.
	 * @param V another vector orthogonal to to <tt>U</tt>; should also be
	 * unit length.
	 */
	public Rotation4D( Vector4D U, Vector4D V )
		throws GeometryException
	{
		/* check for orthogonality of U and V */
		if (U.dot(V) > 1e-16)
			throw new GeometryException("U and V should be orthogonal.");

		/* check for normality */
		U = U.copy();
		V = V.copy();
		U.normalize();
		V.normalize();

		canonical = false;

		construct(xAxis, yAxis);

		Vector4D Zi, Wi;
		double z, w;

		z = - (U.elem[0]*V.elem[3] - V.elem[0]*U.elem[3])
		    / (U.elem[2]*V.elem[3] - V.elem[2]*U.elem[3]);
		w = -1/U.elem[3] * (V.elem[1] + V.elem[2]*z);
		Zi = new Vector4D(0, 1, z, w);
		Wi = U.cross(V, Zi);

		/* note: the M and Minv matrices need to be considered canonical */
		M = new Rotation4D();
		M.setColumn(0, U);
		M.setColumn(1, V);
		M.setColumn(2, Zi);
		M.setColumn(3, Wi);

		Minv = M.inverse();
	}

	/**
	 * Class constructor to generate a 90-degree canonical transformation
	 * around one of the coordinate planes.
	 * @param a1 one of the coordinate axes; for example,
	 * <tt>Rotation4D.xAxis</tt>.
	 * @param a2 the other coordinate axes, to define a plane.
	 */
	public Rotation4D( int a1, int a2 )
		throws GeometryException
	{
		construct(a1, a2);
		canonical = true;
	}

	/**
	 * Constructor helper function for canonical rotations.
	 */
	private void construct( int a1, int a2 )
		throws GeometryException
	{
		int i, j=0;

		canonical = true;

		axes = a1 | a2;

		for (i = 0; i < 4; ++i) {
			if ((axes & (1<<i)) != 0)
				matrix[i][i] = 1;
			else
				inFlumeAxis[j++] = i;
		}
		if (j != 2)
			throw new GeometryException("Need exactly two axis "
				+ "(one plane) to rotate around.");

		matrix[inFlumeAxis[0]][inFlumeAxis[1]] = 
			signRule[inFlumeAxis[0]][inFlumeAxis[1]];
		matrix[inFlumeAxis[1]][inFlumeAxis[0]] = 
			signRule[inFlumeAxis[1]][inFlumeAxis[0]];

		matrix[inFlumeAxis[0]][inFlumeAxis[0]] = 0;
		matrix[inFlumeAxis[1]][inFlumeAxis[1]] = 0;
	}

	/**
	 * Change the angle through which vectors will be rotated.
	 * @param t the new angle to rotate through.
	 */
	public void setAngle( double t )
	{
		theta = t;
		double st = Math.sin(theta), ct = Math.cos(theta);

		if (Math.abs(ct) < 1e-16) ct = 0;
		if (Math.abs(st) < 1e-16) st = 0;

		matrix[inFlumeAxis[0]][inFlumeAxis[1]] = 
			signRule[inFlumeAxis[0]][inFlumeAxis[1]] * st;
		matrix[inFlumeAxis[1]][inFlumeAxis[0]] = 
			signRule[inFlumeAxis[1]][inFlumeAxis[0]] * st;

		matrix[inFlumeAxis[0]][inFlumeAxis[0]] = ct;
		matrix[inFlumeAxis[1]][inFlumeAxis[1]] = ct;

		if (canonical) return;

		/* if we're not dealing with one of the "canonical" rotations (i.e.
		 * those around coordinate planes) then we need to update the matrix
		 * that's actually going to do the rotation. */
		MiRxyM = (Rotation4D)Minv.applyTo(this.applyTo(M));
	}

	/**
	 * Return the angle of rotation.
	 * @return the angle of rotation.
	 */
	public double getAngle()
	{
		return theta;
	}

	/**
	 * Apply this rotation to a vector, allowing specification of angle.
	 * @param t the angle to rotate through.
	 * @param P the vector to rotate.
	 * @return a Vector4D containing the image of <tt>P</tt> under rotation
	 * through <tt>t</tt>.
	 */
	public Vector4D applyTo( double t, Vector4D P )
	{
		this.setAngle(t);
		if (canonical)
			return applyTo(P);
		else
			return MiRxyM.applyTo(P);
	} 

	/**
	 * Compose this rotation with another transformation, allowing
	 * specification of angle.
	 * @param t the angle to rotate through.
	 * @param T the transformation to compose with.
	 * @return a Transform4D representing the composition of this rotation
	 * with <tt>T</tt>.
	 */
	public Transform4D applyTo( double t, Transform4D T )
	{
		this.setAngle(t);
		if (canonical)
			return applyTo(T);
		else
			return MiRxyM.applyTo(T);
	}

	/**
	 * Return an axis specifier given by a string.
	 * @param s a String containing only one of the letters x, y, z, or w,
	 * in upper- or lower-case.
	 * @return an axis specifier (such as <tt>Rotation4D.xAxis</tt>).
	 */
	public static int toAxis( String s )
		throws GeometryException
	{
		if (s.compareToIgnoreCase("x") == 0) return xAxis;
		if (s.compareToIgnoreCase("y") == 0) return yAxis;
		if (s.compareToIgnoreCase("z") == 0) return zAxis;
		if (s.compareToIgnoreCase("w") == 0) return wAxis;
		throw new GeometryException("Invalid axis specification.");
	}

	/**
	 * Return the inverse of this transformation.
	 * @return a new Rotation4D representing the inverse of this rotation.
	 */
	public Rotation4D inverse()
	{
		return transpose();
	}

	/**
	 * Return the transposition of this transformation.
	 * @return a new Rotation4D representing the transformation formed by
	 * the transposition of the matrix of this rotation.
	 */
	public Rotation4D transpose()
	{
		Rotation4D T = new Rotation4D();
		int i;

		for (i = 0; i < 4; ++i)
			T.setRow(i, this.getColumn(i));

		return T;
	}

	/**
	 * Return a copy of this rotation.
	 */
	public Rotation4D copy()
	{
		Rotation4D T = new Rotation4D();
		Transform4D TT = super.copy();
		T.matrix = TT.matrix;
		T.theta = theta;
		return T;
	}
};
