/**
 * The Tesseract class.  Represents a four-dimensional hyperbox.
 */
import java.util.Arrays;

public class Tesseract extends Entity {
	private Vector4D size;
	private Quadric hyperface[];
	private Ellipsonk boundingEllipsonk;

	/**
	 * Class constructor allowing specification of all the dimensions of the
	 * tesseract.
	 * @param s a vector containing the dimensions of the tesseract.
	 */
	public Tesseract( Vector4D s )
	{
		initialize(s);
	}

	/**
	 * Class constructor for a 4-hypercube.
	 * @param s the side length of the tesseract.
	 */
	public Tesseract( double s )
	{
		initialize(new Vector4D(s, s, s, s));
	}

	/**
	 * Class constructor for a standard tesseract.  Each edge will have
	 * length 2.
	 */
	public Tesseract()
	{
		initialize(new Vector4D(2, 2, 2, 2));
	}

	private void initialize( Vector4D s )
	{
		size = s;
		hyperface = new Quadric[4];
		boundingEllipsonk = new Ellipsonk(s);

		for (int i = 0; i < 4; ++i)
			hyperface[i] = new Quadric();

		hyperface[0].coeff[2][0][0][0] = 1;
		hyperface[0].coeff[0][0][0][0] = -s.elem[0]*s.elem[0]/4;

		hyperface[1].coeff[0][2][0][0] = 1;
		hyperface[1].coeff[0][0][0][0] = -s.elem[1]*s.elem[1]/4;

		hyperface[2].coeff[0][0][2][0] = 1;
		hyperface[2].coeff[0][0][0][0] = -s.elem[2]*s.elem[2]/4;

		hyperface[3].coeff[0][0][0][2] = 1;
		hyperface[3].coeff[0][0][0][0] = -s.elem[3]*s.elem[3]/4;
	}

	/**
	 * Collision detector for the tesseract.
	 * @param R the Ray to trace.
	 * @return an array of collisions, or <tt>null</tt> if no collisions
	 * occur.
	 */
private static int foo=0;
	public Collision[] detectCollisions( Ray R )
	{
		int i, j, k;
		Collision all[], C[];

		if (boundingEllipsonk.detectCollisions(R) == null)
			return null;

		all = new Collision[8];

		/* find all collisions with the realms bounding this tesseract.  A
		 * realmic hyperspace is marked "entry" if it corresponds to the
		 * smaller of the two values of t found when solving the quadratic
		 * (i.e., the closer of two parallel realms).  Furthermore, any
		 * collision must be in the bounding box of the tesseract. */
		k = 0;
		for (i = 0; i < 4; ++i) {
			C = hyperface[i].detectCollisions(R);
			if (C == null) continue;

			Arrays.sort(C);
			for (j = 0; j < C.length; ++j) {
				C[j].normal = new Vector4D();
				C[j].normal.elem[i] = C[j].location.elem[i];
				C[j].normal.normalize();
				all[k++] = C[j];
			}

			if (n == 1)
				C[0].entry = true;
			else if (C[1].rayPosition > C[0].rayPosition)
				C[0].entry = true;
			else
				C[1].entry = true;
		}

		/* list the collisions in chronological order */
		Arrays.sort(all, 0, k);

		/* determine whether the last frontal hyperface the ray intersected
		 * comes before the first rear hyperface (in which case, the ray
		 * intersected the tesseract...) */
		boolean gotRear=false;
		int n, lastEntry=-1;
		for (i = 0; i < k; ++i)
			if (all[i].entry) {
				if (gotRear) return null;
				else lastEntry = i;
			} else if (!all[i].entry)
				gotRear = true;
//for (i = 0; i < k; ++i)
	//System.out.println(all[i]);
//System.out.println("\nGot " + k + " collisions.");

		if (lastEntry < 0)
			return null;

		n = lastEntry == all.length-1 ? 1 : 2;
		Collision r[] = new Collision[n];
		r[0] = all[lastEntry];
		if (n == 2)
			r[1] = all[lastEntry+1];
//System.out.println("Entry at " + r[0] + " n = " + r[0].normal);

		/* check for boundedness */
		/* FIXME: this sometimes breaks */
		double S[] = size.toArray();
		for (i = 0; i < n; ++i) {
			double a[] = r[i].location.toArray();
			for (j = 0; j < 4; ++j)
				if (a[j] < -S[j] || a[j] > S[j]) {
					return null;
				}
		}

		return r;
	}

	public String toString()
	{
		return "[ " + size + " @ " + translation + " ]";
	}

	public double getHeight()  { return size.elem[1]; }
	public double getWidth()   { return size.elem[0]; }
	public double getDepth()   { return size.elem[2]; }
	public double getTrength() { return size.elem[3]; }
};
