/**
 * The Transform4D class.  This class represents a linear transformation in
 * tetraspace, on a homogeneous coordinate system.
 */
public class Transform4D {
	/**
	 * The matrix representation of the transformation.
	 */
	public double matrix[][];

	/**
	 * Class constructor which creates an empty matrix.
	 */
	public Transform4D() { matrix = new double[4][4]; }

	/**
	 * Copy the data in vector <tt>V</tt> to column <tt>n</tt> of the
	 * matrix.
	 * @param n the column to update.
	 * @param V the vector to take the column from.
	 */
	public void setColumn( int n, Vector4D V )
	{
		int i;

		for (i = 0; i < 4; ++i)
			matrix[i][n] = V.elem[i];
	}

	/**
	 * Return the <tt>n</tt>th column as a vector.
	 * @param n the index of the column to return.
	 * @return a Vector4D containing the column data.
	 */
	public Vector4D getColumn( int n )
	{
		int i;
		Vector4D V = new Vector4D();

		for (i = 0; i < 4; ++i)
			V.elem[i] = matrix[i][n];

		return V;
	}

	/**
	 * Copy the data in vector <tt>V</tt> to row <tt>n</tt> of the
	 * matrix.
	 * @param n the row to update.
	 * @param V the vector to take the row from.
	 */
	public void setRow( int n, Vector4D V )
	{
		matrix[n] = V.toArray();
	}

	/**
	 * Return the <tt>n</tt>th row as a vector.
	 * @param n the index of the row to return.
	 * @return a Vector4D containing the row data.
	 */
	public Vector4D getRow( int n )
	{
		return new Vector4D(matrix[n]);
	}

	/**
	 * Apply this transformation to a vector.
	 * @param P the Vector4D to apply the transformation to.
	 * @return a Vector4D which is the image of <tt>P</tt> under this
	 * transformation.
	 */
	public Vector4D applyTo( Vector4D P )
	{
		int i, j;
		double q[] = new double[4], p[] = P.toArray();

		for (i = 0; i < 4; ++i)
			for (j = 0; j < 4; ++j)
				q[i] += matrix[i][j] * p[j];

		return new Vector4D(q);
	}

	/**
	 * Apply this transformation to another transformation.
	 * @param T the Transform4D to apply the transformation to.
	 * @return a Transform4D which is the composition of this transformation
	 * with <tt>T</tt>.
	 */
	public Transform4D applyTo( Transform4D T )
	{
		Transform4D S = new Transform4D();
		int i, j, k;

		for (i = 0; i < 4; ++i)
			for (j = 0; j < 4; ++j)
				for (k = 0; k < 4; ++k)
					S.matrix[i][j] += matrix[i][k] * T.matrix[k][j];

		return S;
	}

	/**
	 * Return a string representation of this transformation.  It will take
	 * five lines, each of which is a row vector.  It does not end in a
	 * newline.
	 * @return a String representing this transformation.
	 */
	public String toString()
	{
		String s = "";
		int i, j;

		for (i = 0; i < 4; ++i) {
			if (i == 0)
				s = s + "[ ";
			else
				s = s + "  ";
			s = s + (new Vector4D(matrix[i]))
			      + (i == 3 ? " ]" : "\n");
		}

		return s;
	}

	/**
	 * Return a new copy of this transformation.
	 */
	public Transform4D copy()
	{
		Transform4D T = new Transform4D();

		int i, j;
		for (i = 0; i < 4; ++i)
			for (j = 0; j < 4; ++j)
				T.matrix[i][j] = matrix[i][j];

		return T;
	}
};
