/**
 * The Vector4D class.  This class describes a four-dimensional vector in
 * homogeneous coordinates (though the user can ignore this feature as every
 * vector defaults to being a point).
 */

import java.text.NumberFormat;

public class Vector4D {
	public double elem[];
	public static NumberFormat fmt = NumberFormat.getInstance();

	/**
	 * Class constructor that creates the origin.
	 */
	public Vector4D()
	{
		elem = new double[4];
		fmt.setMaximumFractionDigits(4);
	}

	/**
	 * Class constructor that duplicates an existing vector.
	 * @param V the vector to duplicate.
	 */
	public Vector4D( Vector4D V )
	{
		elem = new double[4];
		for (int i = 0; i < 4; ++i)
			elem[i] = V.elem[i];
		fmt.setMaximumFractionDigits(4);
	}

	/**
	 * Class constructor that allows specification of the coordinates in
	 * standard realmspace.
	 * @param X the <i>x</i>-coordinate.
	 * @param Y the <i>y</i>-coordinate.
	 * @param Z the <i>z</i>-coordinate.
	 */
	public Vector4D( double X, double Y, double Z )
	{
		elem = new double[4];
		elem[0] = X;
		elem[1] = Y;
		elem[2] = Z;
		elem[3] = 0;
		fmt.setMaximumFractionDigits(4);
	}

	/**
	 * Class constructor that allows specification of the coordinates in
	 * tetraspace.
	 * @param X the <i>x</i>-coordinate.
	 * @param Y the <i>y</i>-coordinate.
	 * @param Z the <i>z</i>-coordinate.
	 * @param W the <i>w</i>-coordinate.
	 */
	public Vector4D( double X, double Y, double Z, double W )
	{
		elem = new double[4];
		elem[0] = X;
		elem[1] = Y;
		elem[2] = Z;
		elem[3] = W;
		fmt.setMaximumFractionDigits(4);
	}

	/**
	 * Class constructor that takes values from an array.  The remaining
	 * elements default to the origin's coordinates.
	 * @param a an array containing the elements to take.
	 */
	public Vector4D( double a[] )
	{
		int i;

		for (i = 0; i < 5; ++i)
			if (a.length > i)
				elem[i] = a[i];
			else
				elem[i] = 0;

		fmt.setMaximumFractionDigits(4);
	}

	/**
	 * Return the coordinates of the vector as a five-element array.
	 * @return a five-element array containing the coordinates.
	 */
	public double[] toArray()
	{
		double a[] = new double[4];
		for (int i = 0; i < 4; ++i)
			a[i] = elem[i];
		return a;
	}

	/**
	 * Generate a String describing the vector.
	 * @return a String.
	 */
	public String toString()
	{
		return "( "
			+ fmt.format(elem[0]) + ", "
			+ fmt.format(elem[1]) + ", "
			+ fmt.format(elem[2]) + ", "
			+ fmt.format(elem[3]) + " )";
	}

	/**
	 * Return the L2-norm of the vector.
	 * @return the L2-norm of the vector.
	 */
	public double norm()
	{
		double n = Math.sqrt(elem[0]*elem[0] + elem[1]*elem[1] +
		                     elem[2]*elem[2] + elem[3]*elem[3]);
		return Math.abs(n) < 1e-16 ? 0.0 : n;
	}

	/**
	 * Normalize this vector.
	 */
	public void normalize()
	{
		double n = norm();
		for (int i = 0; i < 4; ++i)
			elem[i] /= n;
	}

	/**
	 * Compute the dot product of this vector with another.
	 * @param V the other vector in the dot product.
	 * @return the dot product.
	 */
	public double dot( Vector4D V )
	{
		double d = 0.0;
		for (int i = 0; i < 4; ++i)
			d += elem[i] * V.elem[i];
		return Math.abs(d) < 1e-16 ? 0.0 : d;
	}

	/**
	 * Compute the determinant of the 3x3 matrix described by <tt>U</tt> and
	 * <tt>V</tt> with column <tt>exclude</tt> excluded.
	 * @param U the vector to use as the second row.
	 * @param V the vector to use as the third row.
	 * @param exclude the index of the column to exclude.
	 * @return the determinant of the matrix generated.
	 */
	private double det33( Vector4D U, Vector4D V, int exclude )
	{
		double d = 0;
		int j[] = { 0, 1, 2 }; /* index to the 3 columns */
		double A[][] = new double[3][];
		double s[] = { 1, -1, 1, -1 };

		switch(exclude) {
		case 0: j[0]++;
		case 1: j[1]++;
		case 2: j[2]++; }

		A[0] = this.toArray();
		A[1] =    U.toArray();
		A[2] =    V.toArray();

		d += s[j[0]] * A[0][j[0]]
		   * (A[1][j[1]]*A[2][j[2]] - A[1][j[2]]*A[2][j[1]]);
		d += s[j[1]] * A[1][j[0]]
		   * (A[0][j[1]]*A[2][j[2]] - A[0][j[2]]*A[2][j[1]]);
		d += s[j[2]] * A[2][j[0]]
		   * (A[0][j[1]]*A[1][j[2]] - A[0][j[2]]*A[1][j[1]]);

		if (Math.abs(d) < 1e-16) d = 0.0;

		return d;
	}

	/**
	 * Return the 4D cross product of three vectors.
	 * @param U the second vector.
	 * @param V the third vector.
	 * @return a new Vector4D containing the 4D cross product.
	 */
	public Vector4D cross( Vector4D U, Vector4D V )
	{
		Vector4D C = new Vector4D();
		double sgn[] = { 1, -1, 1, -1 };

		for (int i = 0; i < 4; ++i)
			C.elem[i] = sgn[i] * det33(U, V, i);

		return C;
	}

	/**
	 * Return a scalar multiple of this vector.
	 * @param c the scalar value to multiply.
	 * @return the scalar multiple of this vector with <tt>c</tt>.
	 */
	public Vector4D scale( double c )
	{
		return new Vector4D(c*elem[0], c*elem[1], c*elem[2], c*elem[3]);
	}

	/**
	 * Return the sum of this vector with another.
	 * @param V the other vector.
	 * @return the sum of this vector and <tt>V</tt>.
	 */
	public Vector4D add( Vector4D V )
	{
		return new Vector4D(elem[0]+V.elem[0], elem[1]+V.elem[1],
		                    elem[2]+V.elem[2], elem[3]+V.elem[3]);
	}

	/**
	 * Return a new copy of this vector.
	 */
	public Vector4D copy()
	{
		return new Vector4D(this);
	}
};
